"""
Module provides tests for the creation of standard tables
"""

import os
import random
from pathlib import Path

from peewee import SqliteDatabase

from mafw.db.db_configurations import default_conf
from mafw.db.db_model import database_proxy
from mafw.db.std_tables import standard_tables
from mafw.decorators import database_required
from mafw.enumerators import LoopType
from mafw.processor import ActiveParameter, Processor, ProcessorList

default_sqlite_pragmas = default_conf['sqlite']['pragmas']


def test_creation_of_std_table():
    database = SqliteDatabase(':memory:', pragmas=default_sqlite_pragmas)
    database_proxy.initialize(database)
    database.create_tables(standard_tables.values())
    for table in standard_tables.values():
        table.init()

    assert len(database.get_tables()) == len(standard_tables)

    TriggerStatus = standard_tables['TriggerStatus']
    assert TriggerStatus.select().count() == 4

    database.close()


def test_creation_of_std_table_via_processor():
    class MyProcessor(Processor):
        n_table = ActiveParameter('n_table', default=-1, help_doc='The number of tables in the database')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

        def process(self):
            self.n_table = len(self.database.get_tables())

    database = SqliteDatabase(':memory:', pragmas=default_sqlite_pragmas)
    database_proxy.initialize(database)

    # when passing a database object, the tables are not created because the processor thinks to be inside a
    # ProcessorList.
    my_processor = MyProcessor(database=database)
    my_processor.execute()

    assert len(database.get_tables()) == 0
    assert my_processor.n_table == 0

    processor_list = ProcessorList(name='A list of processor', database=database)
    processor_list.append(MyProcessor())
    processor_list.execute()
    assert processor_list[0].n_table == 0

    database.close()

    my_processor = MyProcessor(database_conf=default_conf['sqlite'])
    my_processor.execute()
    assert my_processor.n_table >= 2

    processor_list = ProcessorList(name='A list of processor', database_conf=default_conf['sqlite'])
    processor_list.append(MyProcessor())
    processor_list.execute()
    assert processor_list[0].n_table >= 2


def generate_files(path: Path, n_total: int) -> list[Path]:
    files = []
    for i_file in range(1, n_total + 1):
        filename = path / Path(f'file_{i_file:03}.dat')
        files.append(filename)
        filesize = random.randint(1024, 1025)
        with open(filename, 'wb') as fout:
            fout.write(os.urandom(filesize))
    return files


def test_removing_orphan_files(tmp_path, datadir):
    files = generate_files(tmp_path, 10)
    database_conf = default_conf['sqlite'].copy()
    database_conf['URL'] = 'sqlite:///' + str(datadir / 'empty_db.db')

    @database_required
    class FillInOrphanFileProcessor(Processor):
        orphan_files = ActiveParameter(
            'orphan_files', default=[], help_doc='A list of files to be loaded into the orphan table'
        )
        n_files = ActiveParameter('n_files', default=-1, help_doc='Total number of files in the orphan table')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

        def process(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.orphan_files = [dict(filenames=file, checksum=file) for file in self.orphan_files]
            OrphanFile.insert_many(self.orphan_files).execute()
            self.n_files = OrphanFile.select().count()

    @database_required
    class LazyProcessor(Processor):
        initial_n_files = ActiveParameter(
            'initial_n_files', default=-1, help_doc='Total number of files in the orphan table before pruning'
        )
        final_n_files = ActiveParameter(
            'final_n_files', default=-1, help_doc='Total number of files in the orphan table after pruning'
        )

        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper='single', **kwargs)

        def start(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.initial_n_files = OrphanFile.select().count()
            super().start()

        def process(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.final_n_files = OrphanFile.select().count()

    plist = ProcessorList(database_conf=database_conf)

    # this will fill in files to be removed.
    fill_in = FillInOrphanFileProcessor(orphan_files=files)

    # this will count how many files to be removed and then remove them.
    lazy = LazyProcessor()

    plist.append(fill_in)
    plist.append(lazy)

    plist.execute()

    # check that all files were inserted.
    assert fill_in.n_files == len(files)
    # check that all files were still there at the end of the first processor.
    assert lazy.initial_n_files == len(files)
    # check that all files were deleted from the table
    assert lazy.final_n_files == 0
    # check that all files were deleted from the disk
    actual_files = list(tmp_path.glob('*dat'))
    assert len(actual_files) == 0


def test_no_removing_orphan_files(tmp_path, datadir):
    files = generate_files(tmp_path, 10)
    database_conf = default_conf['sqlite'].copy()
    database_conf['URL'] = 'sqlite:///' + str(datadir / 'empty_db.db')

    @database_required
    class FillInOrphanFileProcessor(Processor):
        orphan_files = ActiveParameter(
            'orphan_files', default=[], help_doc='A list of files to be loaded into the orphan table'
        )
        n_files = ActiveParameter('n_files', default=-1, help_doc='Total number of files in the orphan table')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper='single', **kwargs)

        def process(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.orphan_files = [dict(filenames=file, checksum=file) for file in self.orphan_files]
            OrphanFile.insert_many(self.orphan_files).execute()
            self.n_files = OrphanFile.select().count()

    @database_required
    class LazyProcessor(Processor):
        initial_n_files = ActiveParameter(
            'initial_n_files', default=-1, help_doc='Total number of files in the orphan table before pruning'
        )
        final_n_files = ActiveParameter(
            'final_n_files', default=-1, help_doc='Total number of files in the orphan table after pruning'
        )

        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper='single', **kwargs)

        def start(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.initial_n_files = OrphanFile.select().count()
            super().start()

        def process(self):
            OrphanFile = standard_tables.get('OrphanFile')
            self.final_n_files = OrphanFile.select().count()

    plist = ProcessorList(database_conf=database_conf)

    # this will fill in files to be removed.
    fill_in = FillInOrphanFileProcessor(orphan_files=files)

    # this will count how many files to be removed and but do not remove them.
    lazy = LazyProcessor(remove_orphan_files=False)

    plist.append(fill_in)
    plist.append(lazy)

    plist.execute()

    # check that all files were inserted.
    assert fill_in.n_files == len(files)
    # check that all files were still there at the end of the first processor.
    assert lazy.initial_n_files == len(files)
    # check that all files were still there in the table
    assert lazy.final_n_files == len(files)
    # check that all files were still there on the disk
    actual_files = list(tmp_path.glob('*dat'))
    assert len(actual_files) == len(files)



import sys
from typing import Any, Collection

import pytest
from playhouse.db_url import connect

from mafw.decorators import (
    database_required,
    depends_on_optional,
    execution_workflow,
    for_loop,
    orphan_protector,
    processor_depends_on_optional,
    single_loop,
    singleton,
    while_loop,
)
from mafw.enumerators import LoopType
from mafw.mafw_errors import MissingDatabase, MissingOptionalDependency, ProcessorParameterError
from mafw.processor import ActiveParameter, Processor, ensure_parameter_registration


@singleton
class ASingletonClass:
    def __init__(self):
        self.num = 10


def test_singleton():
    first = ASingletonClass()
    second = ASingletonClass()
    assert id(first) == id(second)


def test_database_required(capsys):
    class NormalProcessor(Processor):
        def process(self):
            print('ok')

    p = NormalProcessor(looper=LoopType.SingleLoop)
    p.execute()
    captured = capsys.readouterr()
    assert captured.out == 'ok\n'

    @database_required
    class DBProcessor(Processor):
        def process(self):
            print('db ok')

    p = DBProcessor(looper=LoopType.SingleLoop)
    with pytest.raises(MissingDatabase) as e:
        p.execute()
    assert f'{p.name} requires an active database.' in str(e.value)

    database = connect('sqlite:///:memory:')
    p2 = DBProcessor(looper=LoopType.SingleLoop, database=database)
    p2.execute()
    captured = capsys.readouterr()
    assert 'db ok\n' in captured.out


def test_orphan_protector():
    class RemovalProcessor(Processor):
        pass

    removal = RemovalProcessor(looper=LoopType.SingleLoop)
    assert removal.remove_orphan_files

    @orphan_protector
    class Protector(Processor):
        pass

    protector = Protector(looper=LoopType.SingleLoop)
    assert not protector.remove_orphan_files


def test_ensure_parameter_registration():
    @single_loop
    class MyProcessor(Processor):
        @ensure_parameter_registration
        def a_method(self):
            pass

    @ensure_parameter_registration
    def func():
        pass

    @ensure_parameter_registration
    def another(a, b):
        pass

    my_proc = MyProcessor()
    my_proc.a_method()

    with pytest.raises(ProcessorParameterError):
        func()

    with pytest.raises(ProcessorParameterError):
        another(3, 4)


def test_database_required_wrapping():
    class NotDecoProcessor(Processor):
        """This is a not decorated processor"""

        def __init__(self, *args: Any, **kwargs: Any):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

        def start(self) -> None:
            """This method is not decorated"""
            pass

    assert NotDecoProcessor.__doc__ == 'This is a not decorated processor'
    assert NotDecoProcessor.start.__doc__ == 'This method is not decorated'

    @database_required
    class DBReqProcessor(Processor):
        """This is a DB requiring processor"""

        def start(self) -> None:
            """This is the method that is actually decorated"""
            pass

    assert DBReqProcessor.__doc__ == 'This is a DB requiring processor'
    assert DBReqProcessor.start.__doc__ == 'This is the method that is actually decorated'


def test_workflow_decorators():
    @single_loop
    class SingleProcessor1(Processor):
        pass

    sp1 = SingleProcessor1()
    assert sp1.loop_type == LoopType.SingleLoop

    @execution_workflow(loop_type='single')
    class SingleProcessor2(Processor):
        pass

    sp2 = SingleProcessor2()
    assert sp2.loop_type == LoopType.SingleLoop

    @execution_workflow(loop_type=LoopType.SingleLoop)
    class SingleProcessor3(Processor):
        pass

    sp3 = SingleProcessor3()
    assert sp3.loop_type == LoopType.SingleLoop

    @for_loop
    class ForLoopProcessor1(Processor):
        def get_items(self) -> Collection[Any]:
            return [1]

    flp1 = ForLoopProcessor1()
    assert flp1.loop_type == LoopType.ForLoop

    @execution_workflow(loop_type=LoopType.ForLoop)
    class ForLoopProcessor2(Processor):
        def get_items(self) -> Collection[Any]:
            return [1]

    flp2 = ForLoopProcessor2()
    assert flp2.loop_type == LoopType.ForLoop

    @execution_workflow(loop_type='for_loop')
    class ForLoopProcessor3(Processor):
        def get_items(self) -> Collection[Any]:
            return [1]

    flp3 = ForLoopProcessor3()
    assert flp3.loop_type == LoopType.ForLoop

    @execution_workflow()
    class ForLoopProcessor4(Processor):
        def get_items(self) -> Collection[Any]:
            return [1]

    flp4 = ForLoopProcessor4()
    assert flp4.loop_type == LoopType.ForLoop

    @while_loop
    class WhileLoopProcessor1(Processor):
        def while_condition(self) -> bool:
            return False

    wlp1 = WhileLoopProcessor1()
    assert wlp1.loop_type == LoopType.WhileLoop

    @execution_workflow(loop_type='while_loop')
    class WhileLoopProcessor2(Processor):
        def __init__(self, *args: Any, **kwargs: Any):
            super().__init__(*args, description='a description', **kwargs)

        def while_condition(self) -> bool:
            return False

    wlp2 = WhileLoopProcessor2('wlp2')
    assert wlp2.loop_type == LoopType.WhileLoop
    assert wlp2.description == 'a description'
    assert wlp2.name == 'wlp2'

    @execution_workflow(loop_type=LoopType.WhileLoop)
    class WhileLoopProcessor3(Processor):
        def while_condition(self) -> bool:
            return False

    wlp3 = WhileLoopProcessor3('wlp3', description='test processor')
    assert wlp3.loop_type == LoopType.WhileLoop
    assert wlp3.description == 'test processor'
    assert wlp3.name == 'wlp3'
    assert wlp3.remove_orphan_files is True

    @orphan_protector
    @single_loop
    class SingleProcessor4(Processor):
        pass

    sp4 = SingleProcessor4('sp4', description='doubly decorated')
    assert sp4.loop_type == LoopType.SingleLoop
    assert sp4.remove_orphan_files is False
    assert sp4.name == 'sp4'
    assert sp4.description == 'doubly decorated'


def test_function_depends_on_optional():
    missing_lib = 'missing_lib'
    available_lib = 'sys'

    @depends_on_optional(module_name=available_lib)
    def get_python_version():
        return sys.version_info

    get_python_version()

    # missing lib but all other defaults, raise = False, warn = True
    @depends_on_optional(module_name=missing_lib)
    def failing_python_version_all_defaults():
        return sys.version_info

    with pytest.warns(MissingOptionalDependency, match=missing_lib):
        assert failing_python_version_all_defaults() is None

    # missing lib, raise = True, warn = True
    @depends_on_optional(module_name=missing_lib, raise_ex=True)
    def failing_python_version_raise():
        return sys.version_info

    with pytest.raises(ImportError, match=missing_lib):
        assert failing_python_version_raise() is None

    # missing lib, silent, raise = False, warn = False
    @depends_on_optional(module_name=missing_lib, warn=False)
    def silent_failing():
        return sys.version_info

    assert silent_failing() is None


def test_processor_depends_on_optional(capsys):
    available_libs = 'os;sys'

    @processor_depends_on_optional(available_libs)
    @single_loop
    class GoodProcessor(Processor):
        p1 = ActiveParameter('p1', default=0)

        def process(self) -> None:
            print('GoodProcessor is good')
            print(f'p1 = {self.p1}')

    gp = GoodProcessor()
    gp.execute()
    captured = capsys.readouterr().out
    assert 'GoodProcessor is good\n' in captured
    assert 'p1 = 0' in captured

    gp = GoodProcessor(p1=12)
    gp.execute()
    captured = capsys.readouterr().out
    assert 'GoodProcessor is good\n' in captured
    assert 'p1 = 12\n' in captured
    assert 'p1 = 0\n' not in captured

    missing_libs = 'lib1;lib2'

    with pytest.warns(MissingOptionalDependency, match=missing_libs):
        # raise false, warn true
        @processor_depends_on_optional(missing_libs)
        @single_loop
        class MissingProcessorWarn(Processor):
            p1 = ActiveParameter('p1', default=0)

            def process(self) -> None:
                print('MissingProcessorWarn is missing')
                print(f'p1 = {self.p1}')

    with pytest.raises(ImportError, match=missing_libs):

        @processor_depends_on_optional(missing_libs, raise_ex=True)
        @single_loop
        class MissingProcessorRaise(Processor):
            p1 = ActiveParameter('p1', default=0)

            def process(self) -> None:
                print('MissingProcessorRaise is missing')
                print(f'p1 = {self.p1}')

    @processor_depends_on_optional(missing_libs, raise_ex=False, warn=False)
    @single_loop
    class MissingProcessorSilent(Processor):
        p1 = ActiveParameter('p1', default=0)

        def process(self) -> None:
            print('MissingProcessorSilent is missing')
            print(f'p1 = {self.p1}')

    mps = MissingProcessorSilent(p1=12)
    mps.execute()
    captured = capsys.readouterr().out
    assert captured == ''

    @processor_depends_on_optional(missing_libs, raise_ex=False, warn=False)
    @for_loop
    class MissingForLoopProcessorSilent(Processor):
        p1 = ActiveParameter('p1', default=0)

        def process(self) -> None:
            print('MissingProcessorSilent is missing')
            print(f'p1 = {self.p1}')

    mps = MissingForLoopProcessorSilent(p1=12)
    assert mps.loop_type == LoopType.SingleLoop



import itertools
from pathlib import Path

import pytest
from click.testing import CliRunner
from tomlkit.toml_file import TOMLFile

from mafw import mafw_hookimpl
from mafw.decorators import single_loop
from mafw.mafw_errors import MissingOverloadedMethod, MissingSuperCall
from mafw.plugin_manager import get_plugin_manager
from mafw.processor import Processor
from mafw.scripts import mafw_exe
from mafw.tools.file_tools import file_checksum


def test_running_mafw_without_commands():
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli)
    assert result.output.strip('\n') == 'Use --help to get a quick help on the mafw command.'


def test_help_options():
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['-h'])
    assert 'Usage: mafw [OPTIONS] COMMAND [ARGS]...' in result.output


def test_list_command():
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['list'])
    assert 'Available processors (External Plugins = True)' in result.output

    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['list', '--no-ext-plugin'])
    assert 'Available processors (External Plugins = False)' in result.output


def test_steering_command(tmp_path):
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['steering'])
    assert "Error: Missing argument 'STEERING_FILE'." in result.output

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        filename = 'test.toml'
        result = runner.invoke(mafw_exe.cli, ['steering', filename])
        assert 'A generic steering file has been saved in test.toml.' in result.output
        output_file = Path(Path(td) / Path(filename))
        assert output_file.exists()
        output_file.unlink(missing_ok=True)

        result = runner.invoke(mafw_exe.cli, ['steering', filename, '--show'])
        assert '# MAFw steering file generated ' in result.output
        assert 'A generic steering file has been saved in test.toml.' in result.output
        output_file = Path(Path(td) / Path(filename))
        assert output_file.exists()

        result = runner.invoke(mafw_exe.cli, ['steering', '--db-engine', 'unknown'])
        assert "Error: Invalid value for '--db-engine': " in result.output

        result = runner.invoke(mafw_exe.cli, ['steering', '--db-engine', 'mysql', filename])
        output_file = Path(Path(td) / Path(filename))
        assert output_file.exists()
        with open(output_file) as fp:
            content = fp.read()
        assert 'URL = "mysql://:memory:"' in content


def test_run_command(valid_steering_file, invalid_steering_file):
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['run'])
    assert "Error: Missing argument 'STEERING_FILE'." in result.output

    result = runner.invoke(mafw_exe.cli, ['run', str(valid_steering_file)])
    assert 'Executing AccumulatorProcessor processor' in result.output

    result = runner.invoke(mafw_exe.cli, ['--log-level', 'debug', 'run', str(valid_steering_file)])
    assert 'Executing AccumulatorProcessor processor' in result.output
    assert 'DEBUG    AccumulatorProcessor is initializing' in result.output

    result = runner.invoke(mafw_exe.cli, ['run', str(invalid_steering_file)])
    assert 'ERROR    InvalidSteeringFile: Missing UserInterface' in result.output
    assert 'Traceback (most recent call last)' not in result.output

    result = runner.invoke(mafw_exe.cli, ['-D', 'run', str(invalid_steering_file)])
    assert 'ERROR    Missing section UserInterface' in result.output
    assert 'Traceback (most recent call last)' in result.output


def test_db_group():
    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['db'])
    # expect to read general help from db group
    assert 'Advanced database commands' in result.output


def test_wizard(shared_datadir, tmp_path):
    runner = CliRunner()
    # invoke without arguments
    result = runner.invoke(mafw_exe.cli, ['db', 'wizard'])
    assert "Error: Missing argument 'DATABASE'" in result.output

    # invoke with help
    result = runner.invoke(mafw_exe.cli, ['db', 'wizard', '-h'])
    assert 'mafw db wizard [Options] Database' in result.output

    output_file = tmp_path / 'my_db.py'
    input_database = shared_datadir / 'advanced_db.db'
    result = runner.invoke(mafw_exe.cli, ['db', 'wizard', '-o', str(output_file), str(input_database)])
    # check if output file has been created
    assert output_file.exists()

    # calculate checksum of such file for further use
    checksum = file_checksum(output_file)

    # rerun with no overwrite and cancel
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '--no-overwrite', '-o', str(output_file), str(input_database)], input='c'
    )
    assert output_file.exists()
    assert checksum == file_checksum(output_file)

    # rerun with overwrite
    result = runner.invoke(mafw_exe.cli, ['db', 'wizard', '-o', str(output_file), str(input_database)])
    # check if output file has been created
    assert output_file.exists()
    assert result.exit_code == 0
    # the new file should be different because of the timestamp in the file
    assert checksum != file_checksum(output_file)
    # store the new checksum
    checksum = file_checksum(output_file)

    # rerun with overwrite and backup
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '--no-overwrite', '-o', str(output_file), str(input_database)], input='b'
    )
    assert result.exit_code == 0
    # check if output file has been created
    assert output_file.exists()
    # the new file should be different because of the timestamp in the file
    assert checksum != file_checksum(output_file)
    # find the backup file
    pattern = str(output_file.stem) + '_*' + '.py'
    files = [p for p in Path(tmp_path).glob(pattern)]
    assert len(files) == 1
    bck_file = files[0]
    # check that the bckfile is the original file
    assert checksum == file_checksum(bck_file)

    # try to load the module. other tests are done in test_db_wizard.
    import sys

    sys.path.append(str(output_file.parent))
    try:
        import my_db  # noqa: I001, F401
    except ImportError:
        pytest.fail('Failed to import the generated model.')

    # try to give a wrong engine
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '-e', 'wrongengine', '-o', str(output_file), str(input_database)]
    )
    assert result.exit_code != 0
    assert "Invalid value for '-e'" in result.output

    # try to run with a table subset
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '-t', 'image', '-t', 'resolution', '-o', str(output_file), str(input_database)]
    )
    assert result.exit_code == 0
    # read back the module
    with open(output_file) as in_file:
        module = in_file.read()
    assert 'class Image(MAFwBaseModel):' in module
    assert 'class Resolution(MAFwBaseModel):' in module
    assert 'class Sample(MAFwBaseModel):' in module

    # try to run with a table subset w/o fk
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '-t', 'sample', '-t', 'resolution', '-o', str(output_file), str(input_database)]
    )
    assert result.exit_code == 0
    # read back the module
    with open(output_file) as in_file:
        module = in_file.read()
    assert 'class Image(MAFwBaseModel):' not in module
    assert 'class Resolution(MAFwBaseModel):' in module
    assert 'class Sample(MAFwBaseModel):' in module

    # try to run with a table subset w/o fk and specific view
    result = runner.invoke(
        mafw_exe.cli,
        [
            'db',
            'wizard',
            '-t',
            'sample',
            '-t',
            'resolution',
            '-t',
            'image_view',
            '--with-views',
            '-o',
            str(output_file),
            str(input_database),
        ],
    )
    assert result.exit_code == 0
    # read back the module
    with open(output_file) as in_file:
        module = in_file.read()
    assert 'class Image(MAFwBaseModel):' not in module
    assert 'class Resolution(MAFwBaseModel):' in module
    assert 'class Sample(MAFwBaseModel):' in module
    assert 'class ImageView(MAFwBaseModel):' in module

    # try to run with all tables and views
    result = runner.invoke(mafw_exe.cli, ['db', 'wizard', '--with-views', '-o', str(output_file), str(input_database)])
    assert result.exit_code == 0
    # read back the module
    with open(output_file) as in_file:
        module = in_file.read()
    assert 'class ImageView(MAFwBaseModel):' in module
    assert 'class Resolution(MAFwBaseModel):' in module
    assert 'class Sample(MAFwBaseModel):' in module

    # preserved order
    sorted_class = """
class Image(MAFwBaseModel):
    image_id = AutoField()
    filename = TextField()
    checksum = TextField()
    sample = ForeignKeyField(column_name='sample_id', field='sample_id', model=Sample)
    resolution = ForeignKeyField(column_name='resolution_id', field='resolution_id', model=Resolution)"""

    # try to run with all tables with fields in alphabetical order
    result = runner.invoke(
        mafw_exe.cli, ['db', 'wizard', '--no-preserve-order', '-o', str(output_file), str(input_database)]
    )
    assert result.exit_code == 0
    # read back the module
    with open(output_file) as in_file:
        module = in_file.read()
    assert sorted_class not in module


def test_additional_processors_operation(tmp_path):
    @single_loop
    class Processor1(Processor):
        pass

    @single_loop
    class Processor2(Processor):
        pass

    # create a class container for the plugin hook
    class Container:
        @staticmethod
        @mafw_hookimpl
        def register_processors():
            return [Processor1, Processor2]

    pm = get_plugin_manager()
    pm.register(Container)

    available_processors_list = list(itertools.chain(*pm.hook.register_processors()))
    available_processors: dict[str, type[Processor]] = {p.__name__: p for p in available_processors_list}

    extra_processors = [p.__name__ for p in Container.register_processors()]
    for p in extra_processors:
        assert p in available_processors

    runner = CliRunner()
    result = runner.invoke(mafw_exe.cli, ['list'])
    for p in extra_processors:
        assert p in result.output

    # save a steering file with extra processors
    extra_steering_file = tmp_path / Path('extra_processor.toml')
    result = runner.invoke(mafw_exe.cli, ['steering', str(extra_steering_file)])

    # load the toml file
    toml_dict = TOMLFile(extra_steering_file).read().value
    available_processors_in_toml_file = toml_dict['available_processors']
    assert all([p in available_processors_in_toml_file for p in extra_processors])


def test_warning_catcher(shared_datadir, caplog):
    # create processors emitting warnings:
    class MissingMethodProcessor(Processor):
        # we do not implement the get_items
        pass

    @single_loop
    class MissingSuperProcessor(Processor):
        def start(self) -> None:
            i = 0  # noqa: F841

    class AnotherContainer:
        @staticmethod
        @mafw_hookimpl
        def register_processors():
            return [MissingMethodProcessor, MissingSuperProcessor]

    pm = get_plugin_manager()
    pm.register(AnotherContainer)

    warning_steering_file = shared_datadir / Path('warning_steering.toml')

    # execute the steering file with the two warning emitting processors.
    with pytest.warns((MissingOverloadedMethod, MissingSuperCall)):
        runner = CliRunner()
        result = runner.invoke(mafw_exe.cli, ['run', str(warning_steering_file)])
        assert result.exit_code == 0



import pytest


@pytest.fixture
def valid_steering_file(shared_datadir):
    return shared_datadir / 'valid-steering-file.toml'


@pytest.fixture
def invalid_steering_file(shared_datadir):
    return shared_datadir / 'invalid-steering-file.toml'


@pytest.fixture
def not_existing_steering_file(shared_datadir):
    return shared_datadir / 'invalid-steering-file-xxx.toml'


@pytest.fixture
def valid_db_steering_file(shared_datadir):
    return shared_datadir / 'valid-steering-file-db.toml'



"""
Defines the hook specification decorator bound the MAFw library.
"""

import pluggy

from mafw.processor import Processor
from mafw.ui.abstract_user_interface import UserInterfaceBase

mafw_hookspec = pluggy.HookspecMarker('mafw')


@mafw_hookspec
def register_processors() -> list[Processor]:
    """Register multiple processor classes"""
    return []


@mafw_hookspec
def register_user_interfaces() -> list[UserInterfaceBase]:
    """Register multiple user interfaces"""
    return []



"""
Module provides a simplified script to perform MAFw changelog update using ``auto-changelog``.
It can be used as pre-commit entry point and also in CI.

The basic idea is that this command is invoking the auto-changelog tool to generate a temporary changelog. The
checksum of such a temporary changelog is then compared with the actual one and if it is different, then the current
one is replaced with the newly created.

When committing the changelog update please use mute as commit type, to avoid having a new changelog generated
containing the changelog update commit.

"""

import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import click
from rich import print

from mafw.__about__ import __version__ as mafw_version
from mafw.tools.file_tools import file_checksum


@click.command()
@click.option(
    '-i',
    '--input-file',
    default=Path.cwd() / Path('CHANGELOG.md'),
    type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, allow_dash=False),
    help='The path to the input changelog file. Default CHANGELOG.md',
)
@click.option('-f', '--force-recreate', is_flag=True, default=False, help='Force recreation of CHANGELOG.md')
def update(input_file: click.Path | Path | str, force_recreate: bool) -> int:
    """Execute the auto-changelog program with default configuration for MAFw.

    \f

    :return: 0: if the changelog was successfully updated.
        1: if the changelog was already up to date.
        -1: if an error occurred during the process.
    """
    exe = 'auto-changelog'
    if shutil.which(exe) is None:
        print(
            f'[red]{exe} is not available in this environment. Are you sure, you have installed MAFw with optional dev?'
        )
        return -1

    description = 'MAFw: Modular Analysis Framework'
    latest_version = f'v{mafw_version}'
    tag_prefix = 'v'

    if isinstance(input_file, (click.Path, str)):
        input_file = Path(str(input_file))

    if not input_file.exists():
        print(f'[orange3]No input file {input_file.name} found. Creating a new one.')
        input_file.touch()

    if force_recreate:
        print(f'[orange3]{input_file.name} will be recreated.')
        input_file.unlink()
        input_file.touch()

    original_changelog_cs = file_checksum(input_file)

    with tempfile.TemporaryDirectory() as tmp_dir_name:
        file = Path(tmp_dir_name) / Path(input_file.name)
        args = [
            exe,
            '-o',
            str(file),
            '-d',
            f'"{description}"',
            '--gitlab',
            '-v',
            latest_version,
            '--tag-prefix',
            tag_prefix,
        ]
        subprocess.run(args)
        new_changelog_cs = file_checksum(file)

        if new_changelog_cs != original_changelog_cs:
            shutil.copy(file, input_file)
            print(f'[green]{input_file.name} successfully updated')
            return 0
        else:
            print(f'[cyan]{input_file.name} was already up to date')
            return 2


def main() -> None:
    """Script entry point"""
    ret_val = update(standalone_mode=False)
    sys.exit(ret_val)


if __name__ == '__main__':
    main()



import os
import random
from pathlib import Path
from typing import Any, Iterable

import pytest
from peewee import AutoField, BooleanField, Database, FloatField, ForeignKeyField, SqliteDatabase, TextField
from tomlkit import document, table
from tomlkit.toml_file import TOMLFile

from mafw.db.db_configurations import default_conf
from mafw.db.db_filter import Filter, FilterRegister
from mafw.db.db_model import MAFwBaseModel, database_proxy
from mafw.db.fields import FileChecksumField, FileNameField
from mafw.decorators import single_loop
from mafw.enumerators import LoopType
from mafw.processor import ActiveParameter, Processor


@pytest.fixture
def basic_conf() -> dict:
    doc = document()

    # prepare and add the generic filter table
    tbl = table()
    tbl['new_only'] = True
    tbl['field1'] = 31
    tbl['field3'] = 'something'
    doc.add('GlobalFilter', tbl)

    # prepare and add the processor/model specific filter table
    # first the model table
    mdl_tbl = table()
    mdl_tbl['field1'] = 32
    mdl_tbl['field2'] = 15

    flt_tbl = table()
    flt_tbl.add('MyModel', mdl_tbl)

    proc_tbl = table()
    proc_tbl.add('param1', 12)
    proc_tbl.add('Filter', flt_tbl)

    doc.add('MyProcessor', proc_tbl)

    return doc.value


def test_basic():
    class MeasModel(MAFwBaseModel):
        meas_id = AutoField(primary_key=True)
        sample_name = TextField()
        successful = BooleanField()

    flt = Filter('MyProcessor.Filter.MyModel', sample_name='sample_00*', meas_id=[1, 2, 3], successful=True)
    flt.bind(MeasModel)
    flt.filter()


# noinspection PyUnresolvedReferences
def test_build_filter_from_conf(basic_conf):
    # testing creation of a specific filter overruling generic options and existing
    flt = Filter.from_conf('MyProcessor.Filter.MyModel', basic_conf, default=basic_conf['GlobalFilter'])
    assert flt.field1 == 32
    assert flt.field2 == 15
    assert flt.field3 == 'something'

    # testing creation of a specific filter not existing in the configuration file
    try:
        flt = Filter.from_conf('MyProcessor.Filter.MySecondModel', basic_conf)
    except Exception as e:
        pytest.fail('No error message expected %s', str(e.value))


# noinspection PyUnresolvedReferences
def test_build_filter_from_toml(datadir):
    doc = TOMLFile(datadir / 'basic_conf.toml').read()

    flt = Filter.from_conf('MyProcessor.Filter.MyModel', doc.value)
    assert flt.field1 == 101
    assert flt.field2 == 'ciccia'

    flt = Filter.from_conf('MyProcessor.Filter.MyOtherModel', doc.value)
    assert flt.field1 == -1
    assert flt.field2 == 1

    flt = Filter.from_conf('AnotherProcessor.Filter.MyModel', doc.value)
    assert flt.field1 == 1


def test_processor_and_filter_loading(basic_conf, datadir):
    @single_loop
    class MyProcessor(Processor):
        param1 = ActiveParameter('param1', 144, help_doc='A useless parameter')

    my_proc = MyProcessor(config=basic_conf)
    flt = my_proc.get_filter('MyModel')
    assert my_proc.filter_register.new_only
    assert flt.field1 == 32
    assert flt.field2 == 15

    doc = TOMLFile(datadir / 'basic_conf.toml').read()
    my_proc = MyProcessor(config=doc.value)
    assert len(my_proc.filter_register) == 2
    assert set(list(my_proc.filter_register.keys())) == {'MyModel', 'MyOtherModel'}

    flt = my_proc.get_filter('MyModel')
    assert flt.field2 == 'ciccia'
    assert flt.field1 == 101
    assert flt.field_list == [1, 2, 3]

    flt = my_proc.get_filter('MyOtherModel')
    assert flt.field2 == 1
    assert flt.field1 == -1
    assert not flt.is_bound

    class MyModel(MAFwBaseModel):
        id_ = AutoField(primary_key=True)

    @single_loop
    class AnotherProcessor(Processor):
        pass

    another_proc = AnotherProcessor(config=doc.value)
    assert len(another_proc.filter_register) == 2
    assert set(list(another_proc.filter_register.keys())) == {'MyModel', 'MyOtherModel'}

    flt = another_proc.get_filter('MyModel')
    assert flt.field1 == 1
    assert not flt.is_bound
    flt.bind(MyModel)
    assert flt.is_bound


def test_filter_functionality(datadir):
    doc = TOMLFile(datadir / 'basic_conf.toml').read()

    class AnotherProcessor(Processor):
        pass

    another_proc = AnotherProcessor(config=doc.value, looper=LoopType.SingleLoop)
    for name in another_proc.filter_register:
        # field3 belong only to the GlobalFilter, but all other filter should have
        assert not another_proc.filter_register[name].field3

    class MyModel(MAFwBaseModel):
        id_ = AutoField(primary_key=True)
        sample_name = TextField()

    class MyOtherModel(MAFwBaseModel):
        id_ = AutoField(primary_key=True)
        sample_name = TextField()

    flt1 = another_proc.get_filter('MyModel')
    flt1.bind(MyModel)

    flt2 = another_proc.get_filter('MyOtherModel')
    flt2.bind(MyOtherModel)

    database: Database = SqliteDatabase(':memory:', pragmas=default_conf['sqlite']['pragmas'])
    database.connect()
    database_proxy.initialize(database)
    database.create_tables([MyModel, MyOtherModel], safe=True)
    MyModel.delete().execute()

    data = []
    n = 100
    for i in range(n):
        data.append(dict(sample_name=f'Sample_{i:05}'))
    MyModel.insert_many(data).execute()
    MyOtherModel.insert_many(data).execute()

    assert MyModel.select().count() == n
    assert MyOtherModel.select().count() == n

    filtered = MyModel.select().where(flt1.filter())
    assert filtered.count() == 10

    filtered = MyOtherModel.select().where(flt2.filter())
    assert filtered.count() == len(flt2.id_)
    for f in filtered:
        assert f.id_ in flt2.id_


def test_filter_type_discrimination(datadir):
    doc = TOMLFile(datadir / 'basic_conf.toml').read()

    class MyProcessor(Processor):
        pass

    my_pc = MyProcessor(config=doc.value, looper='single')

    class MyModel(MAFwBaseModel):
        id_ = AutoField(primary_key=True)
        sample_name = TextField()

    flt = my_pc.get_filter('MyModel')
    flt.bind(MyModel)

    assert isinstance(flt.field1, int)
    assert isinstance(flt.field2, str)
    assert isinstance(flt.field_list, list)


@pytest.fixture
def advanced_models() -> tuple:
    class Sample(MAFwBaseModel):
        sample_id = AutoField(primary_key=True, help_text='The sample id primary key')
        sample_name = TextField(help_text='The sample name')

    class Resolution(MAFwBaseModel):
        resolution_id = AutoField(primary_key=True, help_text='The resolution id primary key')
        resolution_value = FloatField(help_text='The resolution in µm')

    class Image(MAFwBaseModel):
        image_id = AutoField(primary_key=True, help_text='The image id primary key')
        filename = FileNameField(help_text='The filename of the image', checksum_field='checksum')
        checksum = FileChecksumField(help_text='The checksum of the input file')
        sample = ForeignKeyField(
            Sample, Sample.sample_id, on_delete='CASCADE', backref='sample', column_name='sample_id'
        )
        resolution = ForeignKeyField(
            Resolution, Resolution.resolution_id, on_delete='CASCADE', backref='resolution', column_name='resolution_id'
        )

    class ProcessedImage(MAFwBaseModel):
        image = ForeignKeyField(
            Image,
            Image.image_id,
            primary_key=True,
            column_name='image_id',
            backref='raw',
            help_text='The image id, foreign key and primary',
            on_delete='CASCADE',
        )
        value = FloatField(default=0)

    return Sample, Resolution, Image, ProcessedImage


def generate_files(path: Path, n_total: int):
    output_files = []
    for i_file in range(1, n_total + 1):
        filename = path / Path(f'file_{i_file:03}.dat')
        filesize = random.randint(1024, 2048)
        with open(filename, 'wb') as fout:
            fout.write(os.urandom(filesize))
        output_files.append(filename)
    return output_files


def test_full_filter_functionality(shared_datadir, advanced_models, datadir):
    db_file = shared_datadir / 'advanced_db.db'
    database: Database = SqliteDatabase(db_file, pragmas=default_conf['sqlite']['pragmas'])
    database.connect()
    database_proxy.initialize(database)

    Sample, Resolution, Image, ProcessedImage = advanced_models
    database.create_tables([Sample, Resolution, Image, ProcessedImage])

    class AdvProcessor(Processor):
        def get_items(self) -> Iterable[Any]:
            """Gets the input lists from a double join and filtering"""
            self.filter_register.bind_all([Sample, Resolution, Image])

            existing_entries = ProcessedImage.select(ProcessedImage.image_id).execute()
            if self.filter_register.new_only:
                existing = ~Image.image_id.in_([i.image_id for i in existing_entries])
            else:
                existing = True

            query = (
                Image.select(Image, Sample, Resolution)
                .join(Sample, on=(Image.sample_id == Sample.sample_id), attr='s')
                .switch(Image)
                .join(Resolution, on=(Image.resolution_id == Resolution.resolution_id), attr='r')
                .where(self.filter_register.filter_all())
                .where(existing)
            )
            return query

        def process(self):
            """Put in the output table only the even images"""
            if self.item.image_id % 2 == 0:
                new_image, _ = ProcessedImage().get_or_create(image_id=self.item.image_id)
                new_image.value = random.randint(0, 100) / random.randint(1, 100)
                new_image.save()

    toml_conf = TOMLFile(datadir / 'adv_db.toml').read().value
    adv_processor = AdvProcessor(database=database, config=toml_conf)
    adv_processor.execute()
    # we expect to have no execution because all output items are already there.
    assert toml_conf['GlobalFilter']['new_only']
    assert adv_processor.n_item == 0
    # check that all filters have the 'just_noise' field
    for flt_name in adv_processor.filter_register:
        assert adv_processor.filter_register[flt_name].just_noise == 'something useless'

    # change conf
    toml_conf['GlobalFilter']['new_only'] = False
    adv_processor = AdvProcessor(database=database, config=toml_conf)
    adv_processor.execute()
    assert adv_processor.n_item == 1
    assert adv_processor.item.image_id == 40

    database.close()


def test_register(datadir, basic_conf):
    # assert default assignment of new_only
    register = FilterRegister()
    assert register.new_only



"""
Module provides a Trigger class and related tools to create triggers in the database via the ORM.
"""

from enum import StrEnum
from typing import Any, Self

import peewee
from peewee import Model

from mafw.mafw_errors import MissingSQLStatement


def and_(*conditions: str) -> str:
    """
    Concatenates conditions with logical AND.

    :param conditions: The condition to join.
    :type conditions: str
    :return: The and-concatenated string of conditions
    :rtype: str
    """
    conditions_l = [f'({c})' for c in conditions]
    return ' AND '.join(conditions_l)


def or_(*conditions: str) -> str:
    """
    Concatenates conditions with logical OR.

    :param conditions: The condition to join.
    :type conditions: str
    :return: The or-concatenated string of conditions.
    :rtype: str
    """
    conditions_l = [f'({c})' for c in conditions]
    return ' OR '.join(conditions_l)


class TriggerWhen(StrEnum):
    """String enumerator for the trigger execution time (Before, After or Instead Of)"""

    Before = 'BEFORE'
    After = 'AFTER'
    Instead = 'INSTEAD OF'


class TriggerAction(StrEnum):
    """String enumerator for the trigger action (Delete, Insert, Update)"""

    Delete = 'DELETE'
    Insert = 'INSERT'
    Update = 'UPDATE'


class Trigger:
    """Trigger template wrapper for use with peewee ORM."""

    def __init__(
        self,
        trigger_name: str,
        trigger_type: tuple[TriggerWhen, TriggerAction],
        source_table: type[Model] | Model | str,
        safe: bool = False,
        for_each_row: bool = False,
        update_columns: list[str] | None = None,
    ):
        """
        Constructor parameters:

        :param trigger_name: The name of this trigger. It needs to be unique!
        :type trigger_name: str
        :param trigger_type: A tuple with :class:`TriggerWhen` and :class:`TriggerAction` to specify on which action
            the trigger should be invoked and if before, after or instead of.
        :type trigger_type: tuple[TriggerWhen, TriggerAction]
        :param source_table: The table originating the trigger. It can be a model class, instance, or also the name of
            the table.
        :type source_table: type[Model] | Model | str
        :param safe: A boolean flag to define if in the trigger creation statement a 'IF NOT EXISTS' clause should be
            included. Defaults to False
        :type safe: bool, Optional
        :param for_each_row: A boolean flag to repeat the script content for each modified row in the table.
            Defaults to False.
        :type for_each_row: bool, Optional
        :param update_columns: A list of column names. When defining a trigger on a table update, it is possible to
            restrict the firing of the trigger to the cases when a subset of all columns have been updated. An column
            is updated also when the new value is equal to the old one. If you want to discriminate this case, use the
            :meth:`add_when` method. Defaults to None.
        :type update_columns: list[str], Optional
        """
        if update_columns is None:
            update_columns = []
        self._create = 'CREATE TRIGGER'
        self.trigger_name = trigger_name
        self.trigger_type = trigger_type
        self._trigger_when, self._trigger_op = self.trigger_type
        self._of_columns = (
            f'OF ({", ".join(update_columns)})' if self._trigger_op == TriggerAction.Update and update_columns else ''
        )
        if isinstance(source_table, (Model, type(Model))):
            # noinspection PyUnresolvedReferences
            self.target_table = source_table._meta.table_name  # type: ignore
        else:
            self.target_table = source_table
        self.if_not_exists = 'IF NOT EXISTS' if safe else ''
        self._for_each_row = 'FOR EACH ROW' if for_each_row else ''

        self._compiled_whens = ''
        self._compiled_sqls = ''
        self._when_list: list[str] = []
        self._sql_list: list[str] = []

    @property
    def trigger_action(self) -> TriggerAction:
        return self._trigger_op

    @trigger_action.setter
    def trigger_action(self, action: TriggerAction) -> None:
        self._trigger_op = action

    @property
    def trigger_when(self) -> TriggerWhen:
        return self._trigger_when

    @trigger_when.setter
    def trigger_when(self, when: TriggerWhen) -> None:
        self._trigger_when = when

    def __setattr__(self, key: Any, value: Any) -> None:
        if key == 'safe':
            self.if_not_exists = 'IF NOT EXISTS' if value else ''
        elif key == 'for_each_row':
            self._for_each_row = 'FOR EACH ROW' if value else ''
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item: str) -> Any:
        if item == 'safe':
            return self.if_not_exists == 'IF NOT EXISTS'
        elif item == 'for_each_row':
            return self._for_each_row == 'FOR EACH ROW'
        else:
            return item

    def add_sql(self, sql: str | peewee.Query) -> Self:
        """
        Add an SQL statement to be executed by the trigger.

        The ``sql`` can be either a string containing the sql statement, or it can be any other peewee Query.

        For example:

        .. code-block:: python

            # assuming you have created a trigger ...

            sql = AnotherTable.insert(
                field1=some_value, field2=another_value
            )
            trigger.add_sql(sql)

        In this way the SQL code is generated with parametric placeholder if needed.

        :param sql: The SQL statement.
        :type sql: str | peewee.Query
        :return: self for easy chaining
        :rtype: Trigger
        """
        if not isinstance(sql, str):
            sql = str(sql)
        sql = sql.strip()
        sql = chr(9) + sql
        if not sql.endswith(';'):
            sql += ';'
        self._sql_list.append(sql)
        return self

    def add_when(self, *conditions: str) -> Self:
        """
        Add conditions to the `when` statements.

        Conditions are logically ANDed.
        To have mixed `OR` and `AND` logic, use the functions :func:`and_` and :func:`or_`.

        :param conditions: Conditions to be added with logical AND
        :type conditions: str
        :return: self for easy chaining
        :rtype: Trigger
        """
        conditions_l = [f'({c.strip()})' for c in conditions]
        self._when_list.append(f'({" AND ".join(conditions_l)})')
        return self

    def create(self) -> str:
        """
        Generates the SQL create statement.

        :return: The trigger creation statement.
        :raise MissingSQLStatement: if no SQL statements are provided.
        """
        if len(self._sql_list) == 0:
            raise MissingSQLStatement('No SQL statements provided')
        self._compiled_whens = f'WHEN {" AND ".join(self._when_list)}' if len(self._when_list) else ''
        self._compiled_sqls = f'{chr(10)}'.join(self._sql_list)
        return (
            f'{self._create} {self.if_not_exists} {self.trigger_name}{chr(10)}'
            f'{self._trigger_when} {self._trigger_op} {self._of_columns} ON {self.target_table}{chr(10)}'
            f'{self._for_each_row} {self._compiled_whens}{chr(10)}'
            f'BEGIN{chr(10)}'
            f'{self._compiled_sqls}'
            f'{chr(10)}END;{chr(10)}'
        )

    def drop(self, safe: bool = True) -> str:
        """
        Generates the SQL drop statement.

        :param safe: If True, add an IF EXIST. Defaults to True.
        :type safe: bool, Optional
        :return: The drop statement
        :rtype: str
        """
        return f'DROP TRIGGER {"IF EXISTS" if safe else ""} {self.trigger_name}'



import pytest

from mafw.enumerators import LoopType


def test_loop_type():
    known_types = ('single', 'for_loop', 'while_loop')
    for t in known_types:
        LoopType(t)

    unknown_type = ('12test34', 'Signle')
    for t in unknown_type:
        with pytest.raises(ValueError, match=t):
            LoopType(t)

    t = LoopType.ForLoop
    nt = LoopType(t)

    assert t == nt



from pathlib import Path
from typing import Any, Collection
from unittest.mock import ANY

import matplotlib.pyplot as plt
import pytest
from peewee import FloatField, TextField
from playhouse.db_url import connect

from mafw.db.db_configurations import default_conf
from mafw.db.db_model import MAFwBaseModel, database_proxy
from mafw.db.std_tables import PlotterOutput
from mafw.decorators import processor_depends_on_optional, single_loop
from mafw.enumerators import ProcessorExitStatus
from mafw.mafw_errors import PlotterMixinNotInitialized
from mafw.processor_library.plotter import (
    CatPlot,
    DisPlot,
    FromDatasetDataRetriever,
    GenericPlotter,
    RelPlot,
    SQLDataRetriever,
)
from mafw.tools.pandas_tools import group_and_aggregate_data_frame

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import seaborn as sns
except ImportError:
    sns = None


pandas_available = pytest.mark.skipif(pd is None, reason='pandas must be available to run this test')
seaborn_available = pytest.mark.skipif(sns is None, reason='seaborn must be available to run this test')


@seaborn_available
def test_backend(caplog):
    pytest.importorskip('tkinter')
    dataframe = sns.load_dataset('penguins')
    plt.switch_backend('Agg')
    sns.relplot(dataframe, x='flipper_length_mm', y='bill_length_mm')
    try:
        plt.switch_backend('TkAgg')
        sns.relplot(dataframe, x='flipper_length_mm', y='bill_length_mm')
    except ImportError:
        pytest.xfail('Missing backend.')

    @single_loop
    class BackendPlotter(GenericPlotter):
        pass

    bp = BackendPlotter(matplotlib_backend='TkAgg')
    assert bp.matplotlib_backend == 'TkAgg'
    bp.start()

    bp.matplotlib_backend = 'Agg'
    bp.start()

    bp.matplotlib_backend = 'Unknown'
    caplog.clear()
    with pytest.raises(ModuleNotFoundError):
        bp.start()

    assert len(caplog.record_tuples) == 1
    assert caplog.record_tuples[0][1] == 50
    assert 'Unknown is not a valid plt backend' in caplog.record_tuples[0][2]


@pandas_available
def test_check_pandas():
    assert pd is not None


@pytest.fixture
def penguin():
    class Penguin(MAFwBaseModel):
        species = TextField(null=True)
        island = TextField(null=True)
        bill_length = FloatField(null=True)
        bill_depth = FloatField(null=True)
        flipper_length = FloatField(null=True)
        body_mass = FloatField(null=True)
        sex = TextField(null=True)

    return Penguin


@pandas_available
def test_super_call_patch_data_frame():
    from mafw.decorators import processor_depends_on_optional, single_loop
    from mafw.mafw_errors import MissingSuperCall
    from mafw.processor_library.plotter import GenericPlotter, SQLDataRetriever

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class GoodPlotter(GenericPlotter):
        def patch_data_frame(self) -> None:
            super().patch_data_frame()
            len(self.data_frame)

    GoodPlotter()

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class MixedGoodPlotter(SQLDataRetriever, GenericPlotter):
        def patch_data_frame(self) -> None:
            super().patch_data_frame()
            len(self.data_frame)

    MixedGoodPlotter()

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class BadPlotter(GenericPlotter):
        def patch_data_frame(self) -> None:
            len(self.data_frame)

    with pytest.warns(MissingSuperCall, match='patch_data_frame'):
        BadPlotter()

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class MixedBadPlotter(SQLDataRetriever, GenericPlotter):
        def patch_data_frame(self) -> None:
            len(self.data_frame)

    with pytest.warns(MissingSuperCall, match='patch_data_frame'):
        MixedBadPlotter()


@pandas_available
def test_plotter_without_database(caplog):
    from mafw.decorators import single_loop
    from mafw.processor_library.plotter import GenericPlotter

    caplog.set_level(30)
    caplog.clear()

    @single_loop
    class NoDBPlotter(GenericPlotter):
        pass

    NoDBPlotter().execute()
    # we expect two warnings, one from the check on output existence, and one from the update output
    assert len(caplog.record_tuples) == 2
    for r in caplog.record_tuples:
        if r[1] == 30:
            assert 'No database connection available' in r[2]
    caplog.clear()

    ndb = NoDBPlotter()
    ndb.filter_register.new_only = False
    ndb.execute()
    # we expect one warning only from the update output
    assert len(caplog.record_tuples) == 1
    for r in caplog.record_tuples:
        if r[1] == 30:
            assert 'No database connection available' in r[2]


@seaborn_available
@pandas_available
def test_mixing_generic_with_sql(datadir, tmp_path, penguin):
    from mafw.decorators import processor_depends_on_optional, single_loop
    from mafw.mafw_errors import PlotterMixinNotInitialized
    from mafw.processor import ActiveParameter
    from mafw.processor_library.plotter import GenericPlotter, SQLDataRetriever

    Penguin: type[MAFwBaseModel] = penguin

    table_name = Penguin._meta.table_name

    cols = ['island', 'species']

    db_file = datadir / Path(r'plotter.db')

    database_conf = default_conf['sqlite']
    database_conf['URL'] = 'sqlite:///' + str(db_file)

    test_df = sns.load_dataset('penguins')

    class PatchedSQLDataRetriever(SQLDataRetriever):
        def patch_data_frame(self) -> None:
            super().patch_data_frame()
            self.data_frame['CAP_ISLAND'] = self.data_frame['island'].str.upper()

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class DataPlotter(PatchedSQLDataRetriever, GenericPlotter):
        param1 = ActiveParameter('param1', default='Test parameter')
        value = ActiveParameter('value', default=12.0)
        data_length = ActiveParameter('data_length', default=0)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.table_name = table_name
            self.required_columns = cols
            self.divider = 10

        def patch_data_frame(self) -> None:
            super().patch_data_frame()
            self.data_frame['CAP_SPECIES'] = self.data_frame['species'].str.upper()

        # the Plotter base has an overloaded process.
        # implement a fake plot to test if it works.
        def plot(self) -> None:
            self.value /= self.divider
            self.data_length = len(self.data_frame.index)

        def save(self) -> None:
            output_file = tmp_path / Path('test.png')
            with open(output_file, 'wt') as f:
                f.write('created by processor')
            self.output_filename_list.append(output_file)

    param1 = 'Another value'
    value = 120.0
    divider = 12

    sql_plotter = DataPlotter(param1=param1, value=value, database_conf=database_conf)

    # setting local processor attribute
    sql_plotter.divider = divider

    # asserting value of mixin attribute initialised in the processor init
    assert sql_plotter.table_name == table_name

    # asserting the value of mixin attribute initialised in the processor init.
    assert sql_plotter.required_columns == cols

    # asserting the value of processor active parameter
    assert sql_plotter.param1 == param1

    # asserting value of processor attribute
    assert sql_plotter.divider == divider

    # asserting initial value of processor attribute
    assert sql_plotter.value == value

    # asserting the initial value of data_length
    assert sql_plotter.data_length == 0

    sql_plotter.execute()
    # assert processor execution was successful
    assert sql_plotter.processor_exit_status == ProcessorExitStatus.Successful

    # assert processor execute was ok
    assert sql_plotter.value == value / divider

    # asserting the real value of data_length
    assert sql_plotter.data_length == len(test_df.index)

    # asserting that the patch data frame implemented into the mixed processor worked
    assert all(sql_plotter.data_frame['CAP_SPECIES'] == test_df['species'].str.upper())

    # asserting that the patch data frame implemented into the derived mixin worked
    assert all(sql_plotter.data_frame['CAP_ISLAND'] == test_df['island'].str.upper())

    # check that the output file was created
    assert (tmp_path / Path('test.png')).exists()

    # remove the file and get ready to repeat the execution
    (tmp_path / Path('test.png')).unlink(missing_ok=True)
    sql_plotter.execute()

    # check that the output file is created
    assert (tmp_path / Path('test.png')).exists()

    # now change the output file
    (tmp_path / Path('test.png')).unlink(missing_ok=True)
    with open(tmp_path / Path('test.png'), 'wt') as f:
        f.write('manually created')

    # run the processor again
    sql_plotter.execute()

    # assert that the output file is existing
    (tmp_path / Path('test.png')).exists()

    # open it and check that is the one created by the processor
    with open(tmp_path / Path('test.png')) as f:
        content = f.read()
    assert 'manually created' not in content
    assert 'created by processor' in content

    # run the processor once again
    # this time the processor is actually doing nothing
    sql_plotter.execute()
    # assert that the output file is existing
    (tmp_path / Path('test.png')).exists()

    @processor_depends_on_optional(module_name='pandas')
    @single_loop
    class NoInitPlotter(PatchedSQLDataRetriever, GenericPlotter):
        param1 = ActiveParameter('param1', default='Test parameter')
        value = ActiveParameter('value', default=12.0)
        data_length = ActiveParameter('data_length', default=0)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.divider = 10

        def patch_data_frame(self) -> None:
            super().patch_data_frame()
            self.data_frame['CAP_SPECIES'] = self.data_frame['species'].str.upper()

        # the Plotter base has an overloaded process.
        # implement a fake plot to test if it works.
        def plot(self) -> None:
            self.value /= self.divider
            self.data_length = len(self.data_frame.index)

    # it must raise an exception
    with pytest.raises(PlotterMixinNotInitialized):
        NoInitPlotter(database_conf=database_conf).execute()

    # it must raise an exception
    with pytest.raises(PlotterMixinNotInitialized):
        NoInitPlotter(database_conf=database_conf, table_name='penguin').execute()

    # again
    with pytest.raises(PlotterMixinNotInitialized):
        NoInitPlotter(required_cols='island', database_conf=database_conf).execute()

    # it should run w/o exception
    nip = NoInitPlotter(
        database_conf=database_conf, table_name='penguin', required_cols=['island', 'species'], where_clause='1'
    )
    nip.execute()

    # it should run w/o exception
    nip = NoInitPlotter(database_conf=database_conf)
    nip.table_name = 'penguin'
    nip.required_columns = ['island', 'species']
    nip.where_clause = ''
    nip.execute()

    # it should run w/o exception
    nip = NoInitPlotter(database_conf=database_conf)
    nip.table_name = 'penguin'
    nip.required_columns = ['island', 'species']
    nip.where_clause = 'where 1'
    nip.execute()


@pandas_available
def test_mixing_generic_with_hdf(datadir, tmp_path, caplog):
    from mafw.processor_library.plotter import GenericPlotter, HDFDataRetriever
    from mafw.tools.pandas_tools import slice_data_frame

    input_file = datadir / Path('plotter.h5')
    input_df = pd.read_hdf(input_file, key='penguins')

    slicing_dict = {'sex': 'Male'}

    sliced_df = slice_data_frame(input_df, slicing_dict)

    test_file = Path('test_hdf.h5')

    @single_loop
    class HDFPlotter(HDFDataRetriever, GenericPlotter):
        def save(self) -> None:
            output_file = tmp_path / test_file
            self.data_frame.to_hdf(output_file, key='penguins')
            self.output_filename_list.append(output_file)

    hdf = HDFPlotter(slicing_dict=slicing_dict, hdf_filename=input_file, key='penguins')
    hdf.execute()

    # assert if the output file has been created
    assert (tmp_path / test_file).exists()
    # assert that the output dataframe is the sliced one
    gen_sliced_df = pd.read_hdf(tmp_path / test_file, key='penguins')
    assert sliced_df.equals(gen_sliced_df)
    (tmp_path / test_file).unlink()

    grouping_cols = ['species', 'island', 'sex']
    aggregate_functions = ['mean', 'std']

    # reuse the same processor instance. add the aggregation functions, but no grouping column
    # we do not expect the aggregation to take place, so the output should be the same as
    # the previous one
    hdf.aggregation_functions = aggregate_functions
    hdf.execute()
    # assert if the output file has been created
    assert (tmp_path / test_file).exists()
    # assert that the output dataframe is the sliced one
    gen_sliced_df = pd.read_hdf(tmp_path / test_file, key='penguins')
    assert sliced_df.equals(gen_sliced_df)
    (tmp_path / test_file).unlink()

    # now add also the grouping cols
    hdf.grouping_columns = grouping_cols
    hdf.execute()
    # assert if the output file has been created
    assert (tmp_path / test_file).exists()
    agg_df = group_and_aggregate_data_frame(sliced_df, grouping_cols, aggregate_functions)
    gen_agg_df = pd.read_hdf(tmp_path / test_file, key='penguins')
    assert agg_df.equals(gen_agg_df)

    # testing without filename nor key in the constructor
    hdf = HDFPlotter()
    with pytest.raises(PlotterMixinNotInitialized):
        hdf.execute()

    # testing with a not exiting filename
    caplog.clear()
    caplog.set_level(30)  # warning
    with pytest.raises(PlotterMixinNotInitialized):
        hdf.hdf_filename = Path('not_existing.h5')
        hdf.execute()
    assert 'not_existing.h5 is not a valid HDF file' in caplog.text

    # testing with a good file, but without key
    with pytest.raises(PlotterMixinNotInitialized):
        hdf.hdf_filename = input_file
        hdf.execute()


@pandas_available
@seaborn_available
def test_mixing_generic_with_dataset(tmp_path):
    valid_dataset_name = 'iris'
    invalid_dataset_name = 'invalid_iris'

    test_output_file = tmp_path / Path('iris.h5')
    input_df = sns.load_dataset(valid_dataset_name)

    @single_loop
    class DataSetPlotter(FromDatasetDataRetriever, GenericPlotter):
        def __init__(self, output_file: str | Path | None = None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.output_file = Path(output_file) if output_file is not None else Path()

        def save(self):
            self.data_frame.to_hdf(self.output_file, key=self.dataset_name)
            self.output_filename_list.append(self.output_file)

    # test with a valid dataset name
    dsp = DataSetPlotter(dataset_name=valid_dataset_name, output_file=test_output_file)
    assert dsp._attributes_valid()
    dsp.execute()
    assert test_output_file.exists()
    gen_df = pd.read_hdf(test_output_file, key=valid_dataset_name)
    assert input_df.equals(gen_df)
    test_output_file.unlink()

    # test without a dataset name
    dsp = DataSetPlotter(output_file=test_output_file)
    assert not dsp._attributes_valid()
    assert dsp.dataset_name == ''
    with pytest.raises(PlotterMixinNotInitialized):
        dsp.execute()
    assert not test_output_file.exists()

    # set an invalid dataset and do the test
    dsp.dataset_name = invalid_dataset_name
    assert not dsp._attributes_valid()
    with pytest.raises(PlotterMixinNotInitialized):
        dsp.execute()
    assert not test_output_file.exists()

    # set a valid dataset and do the test
    dsp.dataset_name = valid_dataset_name
    assert dsp._attributes_valid()
    dsp.execute()
    assert test_output_file.exists()


@pandas_available
@seaborn_available
def test_direct_plot(tmp_path, mocker):
    output_file = tmp_path / Path('relplot.png')

    @single_loop
    @processor_depends_on_optional(module_name='seaborn')
    class SelfDefinedPlotter(FromDatasetDataRetriever, GenericPlotter):
        def __init__(self, output_png: str | Path | None = None, *args, **kwargs):
            super().__init__(dataset_name='penguins', *args, **kwargs)
            self.output_png = Path(output_png) if output_png is not None else Path()

        def plot(self):
            self.facet_grid = sns.relplot(data=self.data_frame, x='flipper_length_mm', y='bill_length_mm', col='sex')
            self.facet_grid.set_axis_labels('Flipper length (mm)', 'Bill length (mm)')

        def save(self) -> None:
            self.facet_grid.figure.savefig(self.output_png)
            self.output_filename_list.append(self.output_png)

    plot_patch = mocker.patch.object(sns, 'relplot', wraps=sns.relplot)
    sdp = SelfDefinedPlotter(output_png=output_file)
    sdp.execute()
    # assert creation of the output file
    assert output_file.exists()
    plot_patch.assert_called_once_with(data=ANY, x='flipper_length_mm', y='bill_length_mm', col='sex')


def adapt_defaults_kwargs(input_kwargs: dict, default_kwargs: dict, expandable_args: list = []) -> dict:
    adapted_kwargs = default_kwargs.copy()

    # expand all expandable in in the input keywords
    for expandable_key in expandable_args:
        if expandable_key in input_kwargs and input_kwargs[expandable_key] is not None:
            for key in input_kwargs[expandable_key]:
                adapted_kwargs[key] = input_kwargs[expandable_key][key]

    # replace the default values with the actual ones.
    for key in default_kwargs:
        if key in input_kwargs:
            adapted_kwargs[key] = input_kwargs[key]

    return adapted_kwargs


@pytest.fixture
def relplot_default_args():
    args = {
        'data': ANY,
        'x': None,
        'y': None,
        'hue': None,
        'row': None,
        'col': None,
        'palette': None,
        'kind': 'scatter',
        'legend': 'auto',
        'facet_kws': None,
    }
    return args


@pandas_available
@seaborn_available
@pytest.mark.parametrize(
    'plotter_param_dict',
    [
        dict(x='flipper_length_mm', y='bill_length_mm', col='sex', hue='species'),
        dict(
            x='flipper_length_mm',
            y='bill_length_mm',
            col='sex',
            hue='species',
            plot_kws={'aspect': 2},
            facet_kws={'legend_out': True},
            palette='pastel',
            legend=True,
        ),
    ],
)
def test_relplot_mixin_param(plotter_param_dict, relplot_default_args, tmp_path, mocker):
    plot_patch = mocker.patch.object(sns, 'relplot', wraps=sns.relplot)

    @single_loop
    @processor_depends_on_optional(module_name='pandas;seaborn')
    class DataSetRelPlotPlotter(FromDatasetDataRetriever, RelPlot, GenericPlotter):
        def __init__(self, output_png: str | Path | None = None, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.output_png = Path(output_png) if output_png is not None else Path()

        def save(self) -> None:
            self.facet_grid.figure.savefig(self.output_png)
            self.output_filename_list.append(self.output_png)

    output_file = tmp_path / Path('relplot_from_dataset.png')
    dsrp = DataSetRelPlotPlotter(output_file, dataset_name='penguins', **plotter_param_dict)
    dsrp.execute()
    assert output_file.exists()

    adapted_kwargs = adapt_defaults_kwargs(plotter_param_dict, relplot_default_args, ['plot_kws'])
    plot_patch.assert_called_once_with(**adapted_kwargs)


@pytest.fixture
def displot_default_args():
    args = {
        'data': ANY,
        'x': None,
        'y': None,
        'hue': None,
        'row': None,
        'col': None,
        'palette': None,
        'kind': 'hist',
        'legend': True,
        'rug': False,
        'rug_kws': None,
        'facet_kws': None,
    }
    return args


@pandas_available
@seaborn_available
@pytest.mark.parametrize(
    'plotter_param_dict',
    [
        dict(x='flipper_length_mm', kind='kde'),
        dict(x='bill_length_mm', kind='hist', rug=True, rug_kws={'clip_on': False}, plot_kws={'kde': True}),
    ],
)
def test_displot_mixin_param(plotter_param_dict, displot_default_args, tmp_path, mocker):
    plot_patch = mocker.patch.object(sns, 'displot', wraps=sns.displot)

    @single_loop
    @processor_depends_on_optional(module_name='pandas;seaborn')
    class DataSetDisPlotPlotter(FromDatasetDataRetriever, DisPlot, GenericPlotter):
        def __init__(self, output_png: str | Path | None = None, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.output_png = Path(output_png) if output_png is not None else Path()

        def save(self) -> None:
            self.facet_grid.figure.savefig(self.output_png)
            self.output_filename_list.append(self.output_png)

    output_file = tmp_path / Path('displot_from_dataset.png')
    dsrp = DataSetDisPlotPlotter(output_file, dataset_name='penguins', **plotter_param_dict)
    dsrp.execute()
    assert output_file.exists()

    adapted_kwargs = adapt_defaults_kwargs(plotter_param_dict, displot_default_args, ['plot_kws'])
    plot_patch.assert_called_once_with(**adapted_kwargs)


@pytest.fixture
def catplot_default_args():
    args = {
        'data': ANY,
        'x': None,
        'y': None,
        'hue': None,
        'row': None,
        'col': None,
        'palette': None,
        'kind': 'strip',
        'legend': 'auto',
        'native_scale': False,
        'facet_kws': None,
    }
    return args


@pandas_available
@seaborn_available
@pytest.mark.parametrize(
    'plotter_param_dict',
    [
        dict(x='age', y='class'),
        dict(x='age', y='class', hue='sex', kind='boxen', plot_kws=dict(legend_out=False)),
        dict(
            x='age',
            y='class',
            hue='sex',
            kind='violin',
            plot_kws=dict(bw_adjust=0.5, cut=0, split=True),
            facet_kws=dict(despine=False),
        ),
    ],
)
def test_catplot_mixin_param(plotter_param_dict, catplot_default_args, tmp_path, mocker):
    plot_patch = mocker.patch.object(sns, 'catplot', wraps=sns.catplot)

    @single_loop
    @processor_depends_on_optional(module_name='pandas;seaborn')
    class DataSetCatPlotPlotter(FromDatasetDataRetriever, CatPlot, GenericPlotter):
        def __init__(self, output_png: str | Path | None = None, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.output_png = Path(output_png) if output_png is not None else Path()

        def save(self) -> None:
            self.facet_grid.figure.savefig(self.output_png)
            self.output_filename_list.append(self.output_png)

    output_file = tmp_path / Path('catplot_from_dataset.png')
    dsrp = DataSetCatPlotPlotter(output_file, dataset_name='titanic', **plotter_param_dict)
    dsrp.execute()
    assert output_file.exists()

    adapted_kwargs = adapt_defaults_kwargs(plotter_param_dict, catplot_default_args, ['plot_kws'])
    plot_patch.assert_called_once_with(**adapted_kwargs)


def test_loop_plotter(tmp_path):
    @processor_depends_on_optional(module_name='pandas;seaborn')
    class LoopPlotter(FromDatasetDataRetriever, DisPlot, GenericPlotter):
        def __init__(self, base_folder: str | Path | None = None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.dataset_name = 'iris'
            self.hue = 'species'
            self.kind = 'kde'
            self.base_output_folder = Path(base_folder) if base_folder is not None else Path()
            self.base_filename = self.dataset_name
            self.output_filename = Path()

        def get_items(self) -> Collection[Any]:
            return ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']

        def in_loop_customization(self) -> None:
            self.x = self.item
            self.output_filename = self.base_output_folder / Path(self.base_filename + '_' + self.item + '.png')

        def customize_plot(self) -> None:
            self.facet_grid.figure.suptitle(f'Distribution for {self.item}')

        def save(self) -> None:
            self.facet_grid.figure.savefig(self.output_filename)
            self.output_filename_list.append(self.output_filename)

    lp = LoopPlotter(base_folder=tmp_path)
    lp.execute()
    assert len(lp.output_filename_list) == len(lp.get_items())
    assert all(f.exists() for f in lp.output_filename_list)


@pandas_available
@seaborn_available
def test_plotter_output(datadir, penguin, tmp_path):
    Penguin: type[MAFwBaseModel] = penguin

    db_file = datadir / Path(r'plotter.db')
    database_conf = default_conf['sqlite']
    database_conf['URL'] = 'sqlite:///' + str(db_file)

    @processor_depends_on_optional(module_name='seaborn')
    @single_loop
    class MyPlotter(SQLDataRetriever, RelPlot, GenericPlotter):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.table_name = Penguin._meta.table_name
            self.required_columns = ['island', 'species', 'sex', 'bill_length', 'bill_depth']
            self.x = 'bill_length'
            self.y = 'bill_depth'
            self.col = 'island'
            self.hue = 'sex'
            self.output_generated = False

        def save(self) -> None:
            output_file = tmp_path / Path('penguin.png')
            self.facet_grid.figure.savefig(output_file)
            self.output_filename_list.append(output_file)
            self.output_generated = True

    plotter = MyPlotter(database_conf=database_conf)
    plotter.execute()

    # check that the output file exists
    assert (tmp_path / Path('penguin.png')).exists()
    # and that the plotter was actually processed till the end.
    assert plotter.output_generated

    database = connect(database_conf['URL'], pragmas=database_conf['pragmas'])
    database_proxy.initialize(database)
    database.connect(reuse_if_open=True)

    # check that the PlotterOutput table exists and that there is MyPlotter
    assert 'plotter_output' in database.get_tables()

    # check that the PlotterOutput has 1 entry.
    assert PlotterOutput.select().count() == 1
    assert PlotterOutput.select().where(PlotterOutput.plotter_name == 'MyPlotter').count() == 1

    plotter = MyPlotter(database_conf=database_conf)
    plotter.execute()

    # the plotter should not have been processed because the output already exists.
    assert not plotter.output_generated

    # remove the output file and re execute
    (tmp_path / Path('penguin.png')).unlink()
    plotter.execute()

    # check that the output file exists
    assert (tmp_path / Path('penguin.png')).exists()
    # and that the plotter was actually processed till the end.
    assert plotter.output_generated

    # remove the entry from the plotter output
    PlotterOutput.delete().where(PlotterOutput.plotter_name == 'MyPlotter').execute()
    assert PlotterOutput.select().where(PlotterOutput.plotter_name == 'MyPlotter').count() == 0
    plotter = MyPlotter(database_conf=database_conf)
    plotter.execute()

    # check that the output file exists
    assert (tmp_path / Path('penguin.png')).exists()
    # and that the plotter was actually processed till the end.
    assert plotter.output_generated
    # assert that the plotter output is populated
    assert PlotterOutput.select().where(PlotterOutput.plotter_name == 'MyPlotter').count() == 1



import time

from mafw.processor import ActiveParameter, Processor
from mafw.ui.abstract_user_interface import UserInterfaceBase
from mafw.ui.rich_user_interface import RichInterface


def test_virtual_inheritance():
    assert issubclass(RichInterface, UserInterfaceBase)


def test_execution_with_single_processor():
    class MyProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=10)
        sleep = ActiveParameter('sleep', default=0.001)

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            time.sleep(self.sleep)

    rich_ui = RichInterface(progress_kws=dict(auto_refresh=True, expand=True))

    p = MyProcessor(user_interface=rich_ui)
    p.execute()



import random
import time
from typing import Any, Collection

import pytest

from mafw.db.db_configurations import default_conf
from mafw.decorators import database_required, single_loop, while_loop
from mafw.enumerators import LoopingStatus, LoopType, ProcessorExitStatus, ProcessorStatus
from mafw.examples.loop_modifier import FindNPrimeNumber, FindPrimeNumberInRange
from mafw.examples.sum_processor import AccumulatorProcessor, GaussAdder
from mafw.mafw_errors import (
    AbortProcessorException,
    MissingDatabase,
    MissingOverloadedMethod,
    MissingSuperCall,
    ProcessorParameterError,
)
from mafw.processor import ActiveParameter, PassiveParameter, Processor, ProcessorList, validate_database_conf


@single_loop
class MyDummyProcessor(Processor):
    active_param: ActiveParameter[int] = ActiveParameter('active_param', default=15, help_doc='An active parameter')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # one compulsory parameter
        self._comp_param: PassiveParameter[int] = PassiveParameter('comp_param', value=5)

        # one optional parameter
        self._opt_param: PassiveParameter[str] = PassiveParameter('opt_param', default='optional')


def test_creation_of_process_parameters():
    with pytest.raises(ProcessorParameterError) as e:
        PassiveParameter('new_param')

    assert 'Processor parameter cannot have both value and default value set to None' == str(e.value)

    for wrong_name in ['12', '1dsfoo-bar', 'bar for']:
        with pytest.raises(ProcessorParameterError) as e:
            PassiveParameter(wrong_name)
        assert 'not a valid python identifier' in str(e)


def test_processor_with_parameters():
    tp = MyDummyProcessor()
    comp_param = tp.get_parameter('comp_param')
    assert comp_param.name == 'comp_param'
    assert comp_param.value == 5
    assert not comp_param.is_optional
    assert comp_param.is_set

    opt_param = tp.get_parameter('opt_param')
    assert opt_param.name == 'opt_param'
    assert opt_param.value == 'optional'
    assert opt_param.is_optional
    assert not opt_param.is_set
    opt_param.value = 'now is set'
    assert opt_param.is_set

    tp.set_parameter_value('opt_param', 'from_proc')
    assert opt_param.value == 'from_proc'
    tp.get_parameter('opt_param').value = 'direct_assignment'
    assert opt_param.value == 'direct_assignment'

    tp.delete_parameter('opt_param')
    with pytest.raises(ProcessorParameterError) as e:
        tp.get_parameter('opt_param')
    assert 'No parameter (opt_param)' in str(e.value)

    assert tp.active_param == tp.get_parameter('active_param').value
    tp.active_param += 1
    assert tp.active_param == tp.get_parameter('active_param').value
    tp.set_parameter_value('active_param', -15)
    assert tp.active_param == tp.get_parameter('active_param').value

    with pytest.raises(ProcessorParameterError) as e:
        tp.get_parameter('missing')
    assert 'No parameter (missing)' in str(e.value)

    class AnotherProcess(Processor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.p1 = PassiveParameter('same_name', default='value')
            self.p2 = PassiveParameter('same_name', default='another_value')

    with pytest.raises(ProcessorParameterError) as e:
        AnotherProcess(looper='single')
    assert 'Duplicated' in str(e.value)

    class ActiveProcessor(Processor):
        param1 = ActiveParameter('param', default=0)
        param2 = ActiveParameter('param', 2)

    with pytest.raises(ProcessorParameterError) as e:
        ActiveProcessor(looper='single')
    assert 'Duplicated' in str(e.value)

    @single_loop
    class ConfigProcessor(Processor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.p1 = PassiveParameter('param1', default='value')
            self.p2 = PassiveParameter('param2', default='another_value')

    cp = ConfigProcessor(param1='1st*', param2='2nd*')
    assert cp.get_parameter('param1').value == '1st*'
    assert cp.get_parameter('param2').value == '2nd*'

    config = {'ConfigProcessor': {'param1': '1st', 'param2': '2nd'}}
    cp = ConfigProcessor(config=config, param2='2nd*')
    assert cp.get_parameter('param1').value == '1st'  # this should remain unchanged.
    assert cp.get_parameter('param2').value == '2nd*'

    config = {'WrongNameProcessor': {'param1': '1st', 'param2': '2nd'}}
    cp = ConfigProcessor(config=config)
    assert cp.get_parameter('param1').value == 'value'
    assert cp.get_parameter('param2').value == 'another_value'


def test_processor_status():
    class MyProcessor(Processor):
        def on_processor_status_change(self, old_status: ProcessorStatus, new_status: ProcessorStatus):
            if new_status == ProcessorStatus.Init:
                assert self.processor_status == ProcessorStatus.Init

    my_p = MyProcessor(looper='single')
    assert my_p.processor_status == ProcessorStatus.Init
    my_p.start()
    assert my_p.processor_status == ProcessorStatus.Start
    my_p.finish()
    assert my_p.processor_status == ProcessorStatus.Finish


def test_processor_looper():
    import random

    from mafw.examples.sum_processor import AccumulatorProcessor

    last_value = random.randint(1, 1000)

    ap = AccumulatorProcessor(last_value=last_value)
    ap.execute()
    assert ap.accumulated_value == last_value * (last_value - 1) / 2


def test_processor_no_loop():
    import random

    from mafw.examples.sum_processor import GaussAdder

    last_value = random.randint(1, 1000)

    gs = GaussAdder(last_value=last_value)
    gs.execute()
    assert gs.sum_value == last_value * (last_value - 1) / 2


def test_adders():
    import random

    from mafw.examples.sum_processor import AccumulatorProcessor, GaussAdder

    last_value = random.randint(1, 1000)
    accumulator = AccumulatorProcessor(last_value=last_value)
    accumulator.execute()
    gauss = GaussAdder(last_value=last_value)
    gauss.execute()
    assert gauss.sum_value == accumulator.accumulated_value


def test_loop_modifier():
    from mafw.examples.loop_modifier import ModifyLoopProcessor

    # generate a random number corresponding to the last item
    last_value = random.randint(10, 1000)

    # get a sample with event to be skipped
    skip_items = random.sample(range(last_value), k=4)

    # find an event to abort after the last skipped one
    max_skip = max(skip_items)
    if max_skip + 1 < last_value:
        abort_item = max_skip + 1
    else:
        abort_item = last_value - 1

    # create the processor and execute it
    mlp = ModifyLoopProcessor(total_item=last_value, items_to_skip=skip_items, item_to_abort=abort_item)
    mlp.execute()

    # compare the recorded skipped items with the list we provided.
    assert mlp.skipped_items == list(sorted(skip_items))

    # check that the last item was the abort item.
    assert mlp.item == abort_item


def test_processor_list():
    # creation and execution of a processor list with no arguments.
    proc_list = ProcessorList()
    proc_list.execute()

    # creation and execution of a processor list with name but no arguments
    proc_list = ProcessorList(name='empty list')
    proc_list.execute()

    # creation and execution of a processor list with name and a list of processors
    last_value = random.randint(1, 1000)
    ga = GaussAdder(last_value=last_value)
    ap = AccumulatorProcessor(last_value=last_value)
    proc_list = ProcessorList(ga, ap, name='two item list')
    proc_list.execute()
    assert ap.accumulated_value == int(last_value * (last_value - 1) / 2)
    assert ga.sum_value == ap.accumulated_value

    # creation and execution of processor list with name and no arguments, but appending processors
    # warning adding two copies of the same processor, will cause the parameter of both to change in the same way.
    # parameters are class attributes!
    last_value += 1
    proc_list2 = ProcessorList(name='items will be appended')
    ga = GaussAdder(last_value=last_value)
    proc_list2.append(ga)
    ap = AccumulatorProcessor(last_value=last_value)
    proc_list2.append(ap)
    proc_list2.execute()
    assert ap.accumulated_value == int(last_value * (last_value - 1) / 2)
    assert ga.sum_value == ap.accumulated_value

    # extending proc_list with proc_list2
    proc_list.extend(proc_list2)
    proc_list.name = 'extended list'
    assert len(proc_list) == 4

    # they were all already executed. the output already exists.
    assert proc_list[0].sum_value == proc_list[1].accumulated_value
    assert proc_list[2].sum_value == proc_list[3].accumulated_value
    assert proc_list[1].accumulated_value != proc_list[3].accumulated_value
    assert proc_list[0].sum_value != proc_list[2].sum_value

    # execute all of them again. this time the last_value will be the same for all because of class instance limitation
    proc_list.execute()
    assert proc_list[0].sum_value == proc_list[1].accumulated_value
    assert proc_list[2].sum_value == proc_list[3].accumulated_value
    assert proc_list[1].accumulated_value == proc_list[3].accumulated_value
    assert proc_list[0].sum_value == proc_list[2].sum_value

    # processor list nesting
    last_value += 1
    nest_list = ProcessorList(name='nested')

    list1 = ProcessorList(name='list1')
    list1.append(GaussAdder(last_value=last_value))
    list1.append(AccumulatorProcessor(last_value=last_value))

    list2 = ProcessorList(name='list2')
    list2.extend([GaussAdder(last_value=last_value), AccumulatorProcessor(last_value=last_value)])

    nest_list.append(list1)
    nest_list.append(list2)
    assert len(nest_list) == 2

    nest_list.execute()
    assert nest_list[0][0].sum_value == nest_list[1][0].sum_value
    assert nest_list[0][0].sum_value == int(last_value * (last_value - 1) / 2)

    # checking resource distribution
    list1 = ProcessorList(name='list1')
    list1.append(GaussAdder(last_value=last_value))
    list1.append(AccumulatorProcessor(last_value=last_value))
    list1.execute()

    # list1 should have a timer duration, and each of its members should have a timer with the same duration
    assert list1.timer.duration != 0
    assert all([list1.timer.duration == item.timer.duration for item in list1])


def test_processor_list_abort(capsys):
    # test AbortProcessorException
    class GoodProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100, help_doc='The n of the loop')
        sleep_time = ActiveParameter('sleep_time', default=0.01, help_doc='So much work')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            # pretend to do something, but actually sleep
            time.sleep(self.sleep_time)

        def finish(self):
            super().finish()
            print(f'{self.name} just finished with status: {self.processor_exit_status.name}')

    class BadProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100, help_doc='The n of the loop')
        sleep_time = ActiveParameter('sleep_time', default=0.01, help_doc='So much work')
        im_bad = ActiveParameter('im_bad', default=50, help_doc='I will crash it!')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            if self.item == self.im_bad:
                self.looping_status = LoopingStatus.Abort
                return
            # let me do my job
            time.sleep(self.sleep_time)

        def finish(self):
            super().finish()
            print(f'{self.name} just finished with status: {self.processor_exit_status.name}')

    proc_list = ProcessorList(name='with exception')
    proc_list.extend([GoodProcessor(), BadProcessor(), GoodProcessor()])
    with pytest.raises(AbortProcessorException):
        proc_list.execute()

    output = capsys.readouterr()
    assert 'GoodProcessor just finished with status: Successful' in output.out.split('\n')
    assert 'BadProcessor just finished with status: Aborted' in output.out.split('\n')
    assert proc_list.processor_exit_status == ProcessorExitStatus.Aborted
    assert proc_list[0].processor_exit_status == ProcessorExitStatus.Successful
    assert proc_list[1].processor_exit_status == ProcessorExitStatus.Aborted
    assert proc_list[2].processor_status == ProcessorStatus.Init


def test_active_parameter():
    from mafw.processor import ActiveParameter, Processor

    @single_loop
    class MyProcess(Processor):
        my_param = ActiveParameter('my_param', default=10)

    my_proc = MyProcess(my_param=12)
    print(my_proc.my_param)  # we expect 12
    assert my_proc.my_param == 12

    second_proc = MyProcess(my_param=15)
    print(second_proc.my_param)  # we expect 15
    assert second_proc.my_param == 15

    print(my_proc.my_param)  # we would expect 12, but we get 15
    assert my_proc.my_param == 15


def test_passive_parameter():
    from mafw.processor import PassiveParameter, Processor

    class MyProcess(Processor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, looper='single', **kwargs)
            self.my_param = PassiveParameter('my_param', default=10)

    my_proc = MyProcess(my_param=12)
    print(my_proc.my_param.value)  # we expect 12
    assert my_proc.my_param.value == 12

    second_proc = MyProcess(my_param=15)
    assert second_proc.my_param.value == 15
    print(second_proc.my_param.value)  # we expect 15

    print(my_proc.my_param.value)  # we expect 12 and we get 12!
    assert my_proc.my_param.value == 12


def test_parameter_keyword():
    class MyProcessor(Processor):
        active_param = ActiveParameter('active', default=0, help_doc='An active parameter')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.passive_param = PassiveParameter('passive', default='I am a string', help_doc='A string')

    my_p = MyProcessor(active=100, passive='a better string', looper='single')

    print(my_p.active_param)  # we get 100
    assert my_p.active_param == 100

    print(my_p.passive_param.value)  # we get 'a better string'
    assert my_p.passive_param.value == 'a better string'


def test_parameter_configuration():
    @single_loop
    class ConfigProcessor(Processor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.p1 = PassiveParameter('param1', default='value')
            self.p2 = PassiveParameter('param2', default='another_value')

    cp = ConfigProcessor(config=dict(param1='new_value', param2='better_value', param3='do not exists'))
    assert cp.get_parameter('param1').value == 'new_value'
    assert cp.get_parameter('param2').value == 'better_value'
    dumped_config = cp.dump_parameter_configuration(option=2)
    assert dumped_config == dict(param1='new_value', param2='better_value')
    dumped_config = cp.dump_parameter_configuration(option=20)
    assert dumped_config == dict(param1='new_value', param2='better_value')

    config = {'ConfigProcessor': {'param1': '1st', 'param2': '2nd'}}
    cp = ConfigProcessor(config=config)
    assert cp.get_parameter('param1').value == '1st'
    assert cp.get_parameter('param2').value == '2nd'
    dumped_config = cp.dump_parameter_configuration()
    assert config == dumped_config


def test_processor_list_exit_status():
    class SimpleProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100)

        def start(self):
            super().start()
            print(f'{self.name} is working')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

    class FailingProcessor(Processor):
        n_loop = ActiveParameter('n_loop', default=100)
        i_fail = ActiveParameter('i_fail', default=10)

        def start(self):
            super().start()
            print(f'{self.name} is working')

        def get_items(self) -> list[int]:
            return list(range(self.n_loop))

        def process(self):
            if self.i_item == self.i_fail:
                self.processor_exit_status = ProcessorExitStatus.Aborted

    processor_list = ProcessorList(name='A processor list')
    processor_list.append(FailingProcessor(n_loop=100, i_fail=50))
    processor_list.append(SimpleProcessor(n_loop=100))
    with pytest.raises(AbortProcessorException) as e:
        processor_list.execute()
    assert 'FailingProcessor' in str(e.value)
    assert processor_list[0].i_item == 99
    assert processor_list[1].processor_status == ProcessorStatus.Init


def test_unique_id():
    @single_loop
    class MyProcessor(Processor):
        pass

    @single_loop
    class MyProcessor2(Processor):
        pass

    p0 = MyProcessor()
    p1 = MyProcessor(name='ATestProcessor')
    p2 = MyProcessor(name='ATestProcessor')
    p3 = MyProcessor2(name='ATestProcessor')
    assert p0.unique_name != p1.unique_name != p2.unique_name != p3.unique_id


def test_validate_db_conf():
    # when taken from the module, the configuration is of type 2.
    type2 = default_conf['sqlite']
    validated_conf2 = validate_database_conf(type2)
    assert type2 == validated_conf2

    # transform the conf in a type 1.
    type1 = dict(DBConfiguration=type2)
    type1['OtherField'] = 5

    validated_conf1 = validate_database_conf(type1)
    assert 'OtherField' in validated_conf1
    assert 'DBConfiguration' in validated_conf1


def test_processor_with_db_conf(capsys):
    # processor with no db, but with db_conf
    @single_loop
    class DBProcessor(Processor):
        pass

    # check of setting of the db configuration
    db_proc = DBProcessor(database_conf=default_conf['sqlite'])
    assert db_proc._database is None
    with pytest.raises(MissingDatabase):
        db_proc.database
    assert db_proc._database_conf == default_conf['sqlite']
    db_proc.execute()
    assert db_proc.database is not None
    assert db_proc.database.is_closed()

    type1 = {'DBConfiguration': default_conf['sqlite'], 'OtherKey': 5}
    db_proc = DBProcessor(database_conf=type1)
    assert db_proc._database is None
    with pytest.raises(MissingDatabase):
        db_proc.database
    assert db_proc._database_conf == type1
    db_proc.execute()
    assert db_proc.database is not None
    assert db_proc.database.is_closed()

    invalid = {'URLS': 'sqlite:///:memory:'}  # URLS instead of URL
    db_proc = DBProcessor(database_conf=invalid)
    assert db_proc._database is None
    with pytest.raises(MissingDatabase):
        db_proc.database
    assert db_proc._database_conf is None
    db_proc.execute()
    assert db_proc._database is None
    with pytest.raises(MissingDatabase):
        db_proc.database

    @database_required
    @single_loop
    class ReqDBProcessor(Processor):
        def start(self):
            super().start()
            print('starting')

    db_proc = ReqDBProcessor(database_conf=default_conf['sqlite'])
    db_proc.execute()
    captured = capsys.readouterr()
    assert 'starting\n' in captured.out

    db_proc = ReqDBProcessor(database_conf=invalid)
    with pytest.raises(MissingDatabase) as e:
        db_proc.execute()
    assert f'{db_proc.name} requires an active database.' in str(e.value)


def test_method_overload():
    class SingleProcessor(Processor):
        def __init__(self, *args: Any, **kwargs: Any):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

    SingleProcessor()

    class ForLoopProcessor(Processor):
        pass

    with pytest.warns(MissingOverloadedMethod, match='get_items'):
        ForLoopProcessor()

    class ProperForLoopProcessor(Processor):
        def get_items(self) -> Collection[Any]:
            return [12]

    ProperForLoopProcessor()

    @while_loop
    class WhileLoopProcessor(Processor):
        pass

    with pytest.warns(MissingOverloadedMethod, match='while_condition'):
        WhileLoopProcessor()

    @while_loop
    class ProperWhileLoopProcessor(Processor):
        def while_condition(self) -> bool:
            return False

    ProperWhileLoopProcessor()


def test_super_method_call():
    class SingleProcessor(Processor):
        pass

    # no method overload, no need to call super
    SingleProcessor(looper='single')

    # properly overloaded, no warning
    @single_loop
    class ProperlyOverloadedProcessor(Processor):
        def start(self):
            """docstring"""
            # random comment
            super().start()
            i = 10 + 2
            i = i * 2

    ProperlyOverloadedProcessor()

    # Wrongly overloaded, warning!
    class WronglyOverloadedProcessor(Processor):
        def __init__(self, *args: Any, **kwargs: Any):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

        def start(self) -> None:
            i = 10 + 2
            i = i / 2

        def finish(self) -> None:
            """It does not matter when you call the super."""
            # we do something before
            i = 12
            # we call the super
            super().finish()
            # we do something else
            i = i - 1

    with pytest.warns(MissingSuperCall, match='start'):
        WronglyOverloadedProcessor()

    # Wrongly overloaded, warning!
    # super is commented out in start and finish is just pass
    class WronglyOverloadedProcessor2(Processor):
        def __init__(self, *args: Any, **kwargs: Any):
            super().__init__(*args, looper=LoopType.SingleLoop, **kwargs)

        def start(self) -> None:
            # super().start()
            i = 10 + 2
            i = i - 3

        def finish(self) -> None:
            pass

    with pytest.warns(MissingSuperCall, match='[start|finish]'):
        WronglyOverloadedProcessor2()


def test_prime_processors():
    start_from = 20
    stop_at = 50
    prime_num_to_find = 7

    prime_in_range = FindPrimeNumberInRange(start_from=start_from, stop_at=stop_at)
    n_prime = FindNPrimeNumber(start_from=start_from, prime_num_to_find=prime_num_to_find)

    prime_in_range.execute()
    n_prime.execute()

    assert len(prime_in_range.prime_num_found) == len(n_prime.prime_num_found)
    assert len(prime_in_range.prime_num_found) == prime_num_to_find



"""
Module implements a generic plotter processor with a mixin structure to generate seaborn plots.

The basic idea is to have a :class:`basic processor class <.GenericPlotter>` featuring a modified
:meth:`~.GenericPlotter.process` method where a skeleton of standard operations required to generate a graphical
representation of a dataset is provided.

The user has the possibility to compose the :class:`~.GenericPlotter` by mixing it with one :class:`~.DataRetriever`
and a :class:`~.FigurePlotter`. The former will allow to retrieve a data frame from a standard source (:class:`HDF
file <.HDFDataRetriever>` or :class:`SQL table <.SQLDataRetriever>`) while the latter will allow to generate seaborn
figure level graphs (:class:`relational plots <.RelPlot>`, :class:`distribution plots <.DisPlot>` and :class:`categorical plots
<.CatPlot>`) without having to write a single line of code.

"""

import logging
import re
import typing
from collections.abc import Callable, Iterable, Mapping, MutableMapping, Sequence
from pathlib import Path
from typing import Any, Protocol, TypeAlias

import peewee

from mafw.db.std_tables import PlotterOutput
from mafw.decorators import processor_depends_on_optional
from mafw.enumerators import LoopingStatus
from mafw.mafw_errors import PlotterMixinNotInitialized
from mafw.processor import Processor, ProcessorMeta
from mafw.tools.file_tools import file_checksum
from mafw.tools.pandas_tools import group_and_aggregate_data_frame, slice_data_frame

try:
    import matplotlib.colors
    import matplotlib.pyplot as plt
    from matplotlib.typing import ColorType

    _has_pyplot = True
except ImportError:
    _has_pyplot = False
try:
    import pandas as pd

    _has_pandas = True
except ImportError:
    _has_pandas = False

try:
    import seaborn as sns
    from seaborn._core.typing import ColumnName

    # noinspection PyProtectedMember
    _Palette: TypeAlias = str | Sequence[ColorType] | Mapping[Any, ColorType]

    _has_seaborn = True
except ImportError:
    _has_seaborn = False

log = logging.getLogger(__name__)


class PlotterMeta(type(Protocol), ProcessorMeta):  # type: ignore[misc]
    """Metaclass for the plotter mixed classes"""

    pass


class DataRetriever(Protocol):
    """Base mixin class to retrieve a data frame from an external source"""

    data_frame: pd.DataFrame
    """The dataframe instance. It will be filled for the main class"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # leave it here, otherwise the Protocol init will not call the main class init.
        # not sure why this is happening, but it costs nothing to have it here.
        super().__init__(*args, **kwargs)

    def get_data_frame(self) -> None:
        """The mixin implementation of the shared method with the base class"""
        pass

    def patch_data_frame(self) -> None:
        """The mixin implementation of the shared method with the base class"""
        pass

    def _attributes_valid(self) -> bool:
        return True


class FromDatasetDataRetriever(DataRetriever):
    """
    A data retriever to get a dataframe from a seaborn dataset
    """

    def __init__(self, dataset_name: str | None = None, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.dataset_name = dataset_name if dataset_name is not None else ''

    def _attributes_valid(self) -> bool:
        """Checks if the attributes of the mixin are all valid"""
        if not _has_seaborn or self.dataset_name == '':
            return False

        return self.dataset_name in sns.get_dataset_names()

    def get_data_frame(self) -> None:
        """Gets the data frame from the standard seaborn datasets"""
        if not self._attributes_valid():
            raise PlotterMixinNotInitialized()
        self.data_frame = sns.load_dataset(self.dataset_name)

    def patch_data_frame(self) -> None:
        super().patch_data_frame()  # type: ignore[safe-super]


class SQLDataRetriever(DataRetriever):
    """
    A specialised data retriever to get a data frame from a database table.

    The idea is to implement an interface to the pandas ``read_sql``. The user has to provide the :attr:`table name
    <.table_name>`, the :attr:`the list of required columns <.required_columns>` and an optional :attr:`where clause
    <.where_clause>`.
    """

    database: peewee.Database
    """The database instance. It comes from the main class"""

    def __init__(
        self,
        table_name: str | None = None,
        required_cols: Iterable[str] | str | None = None,
        where_clause: str | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Constructor parameters:

        :param table_name: The name of the table from where to get the data
        :type table_name: str, Optional
        :param required_cols: A list of columns to be selected from the table and transferred as column in the dataframe.
        :type required_cols:  Iterable[str] | str | None, Optional
        :param where_clause: The where clause used in the select SQL statement. If None is provided, then all rows will
            be selected.
        :type where_clause: str, Optional
        """
        super().__init__(*args, **kwargs)
        self.table_name: str
        """The table from where the data should be taken."""
        if table_name is None:
            self.table_name = ''
        else:
            self.table_name = table_name

        self.required_columns: Iterable[str]
        """
        The iterable of columns.

        Those are the column names to be selected from the :attr:`~.table_name` and included in the dataframe. 
        """
        if required_cols is None:
            self.required_columns = []
        elif isinstance(required_cols, str):
            self.required_columns = [required_cols]
        else:
            self.required_columns = required_cols

        self.where_clause: str
        """The where clause of the SQL statement"""

        if where_clause is None:
            self.where_clause = '1'
        else:
            self.where_clause = where_clause

    def get_data_frame(self) -> None:
        """
        Retrieve the dataframe from a database table.

        :raise PlotterMixinNotInitialized: If some of the required attributes are missing.
        """
        if not self._attributes_valid():
            raise PlotterMixinNotInitialized

        if isinstance(self.required_columns, str):
            self.required_columns = [self.required_columns]

        if self.where_clause == '':
            where_clause = ' 1 '
        else:
            where_clause = self.where_clause.strip()
            m = re.match('where', where_clause, re.I)
            if m:
                where_clause = where_clause.replace(m[0], '').strip()

        # todo:
        #   sqlite does not allow to have column and table names parametrized.
        #   so we need to concatenate the strings.
        #   see https://www.sqlite.org/cintro.html#binding_parameters_and_reusing_prepared_statements
        #   we should actually do a check for SQL injection for those elements, but I have no idea at the moment how
        #   this could be implemented.
        sql = f'SELECT {", ".join(self.required_columns)} FROM {self.table_name} WHERE ?'
        params = (where_clause,)

        data_frame = pd.read_sql(sql, con=self.database.connection(), params=params)  # type: ignore[no-untyped-call]

        self.data_frame = data_frame

    def _attributes_valid(self) -> bool:
        """Check if all required parameters are provided and valid."""
        if self.table_name == '':
            return False
        if not self.required_columns:
            return False
        return True

    def patch_data_frame(self) -> None:
        """
        Patch the data frame.

        In the basic implementation of the SQLDataRetriever the patch data frame is just invoking the super method to
        be sure that the main class patch data frame is invoked.
        """
        super().patch_data_frame()  # type: ignore[safe-super]


class HDFDataRetriever(DataRetriever):
    """
    Retrieve a data frame from a HDF file

    This data retriever is getting a dataframe from a HDF file provided the filename and the object key.
    """

    def __init__(
        self, hdf_filename: str | Path | None = None, key: str | None = None, *args: Any, **kwargs: Any
    ) -> None:
        """
        Constructor parameters:

        :param hdf_filename: The filename of the HDF file
        :type hdf_filename: str | Path, Optional
        :param key: The key of the HDF store with the dataframe
        :type key: str, Optional
        """
        super().__init__(*args, **kwargs)
        self.hdf_filename: Path
        if hdf_filename is None:
            self.hdf_filename = Path()
        else:
            self.hdf_filename = Path(hdf_filename)

        self.key: str
        if key is None:
            self.key = ''
        else:
            self.key = key

    def get_data_frame(self) -> None:
        """
        Retrieve the dataframe from a HDF file

        :raise PlotterMixinNotInitialized: if some of the required attributes are not initialised or invalid.
        """
        if not self._attributes_valid():
            raise PlotterMixinNotInitialized

        self.data_frame = typing.cast(pd.DataFrame, pd.read_hdf(self.hdf_filename, self.key))

    def patch_data_frame(self) -> None:
        super().patch_data_frame()  # type: ignore[safe-super]

    def _attributes_valid(self) -> bool:
        if self.hdf_filename == Path():
            return False
        elif not self.hdf_filename.is_file():
            # hdf is not a file or it does not exist.
            log.warning('%s is not a valid HDF file' % self.hdf_filename)
            return False

        if self.key == '':
            return False

        return True


class FigurePlotter(Protocol):
    """Base mixin class to generate a seaborn Figure level plot"""

    data_frame: pd.DataFrame
    """The dataframe instance shared with the main class"""

    facet_grid: sns.FacetGrid
    """The facet grid instance shared with the main class"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def plot(self) -> None:
        pass

    def _attributes_valid(self) -> bool:
        return True


class RelPlot(FigurePlotter):
    """
    The relational plot mixin.

    This mixin will produce either a scatter or a line figure level plot.

    The full documentation of the relplot object can be read at `this link <https://seaborn.pydata.org/generated/seaborn.relplot.html>`_.
    """

    def __init__(
        self,
        x: ColumnName | Iterable[float | complex | int] | None = None,
        y: ColumnName | Iterable[float | complex | int] | None = None,
        hue: ColumnName | Iterable[float | complex | int] | None = None,
        row: ColumnName | Iterable[float | complex | int] | None = None,
        col: ColumnName | Iterable[float | complex | int] | None = None,
        palette: _Palette | matplotlib.colors.Colormap | None = None,
        kind: typing.Literal['scatter', 'line'] = 'scatter',
        legend: typing.Literal['auto', 'brief', 'full'] | bool = 'auto',
        plot_kws: Mapping[str, Any] | None = None,
        facet_kws: dict[str, Any] | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Constructor parameters:

        :param x: The name of the x variable or an iterable containing the x values.
        :type x: str | Iterable, Optional
        :param y: The name of the y variable or an iterable containing the y values.
        :type y: str | Iterable, Optional
        :param hue: The name of the hue variable or an iterable containing the hue values.
        :type hue: str | Iterable, Optional
        :param row: The name of the row category or an iterable containing the row values.
        :type row: str | Iterable, Optional
        :param col: The name of the column category or an iterable containing the column values.
        :type col: str | Iterable, Optional
        :param palette: The colour palette to be used.
        :type palette: str | Colormap, Optional
        :param kind: The type of relational plot (scatter or line). Defaults to scatter.
        :type kind: str, Optional
        :param legend: How to draw the legend. If “brief”, numeric hue and size variables will be represented with a
            sample of evenly spaced values. If “full”, every group will get an entry in the legend.
            If “auto”, choose between brief or full representation based on number of levels.
            If False, no legend data is added and no legend is drawn. Defaults to auto.
        :type legend: str | bool, Optional
        :param plot_kws: A dictionary like list of keywords passed to the underlying `seaborn.relplot
            <https://seaborn.pydata.org/generated/seaborn.relplot.html#seaborn.relplot>`_.
        :type plot_kws: dict[str, Any], Optional
        :param facet_kws: A dictionary like list of keywords passed to the underlying
            `seaborn.FacetGrid <https://seaborn.pydata.org/generated/seaborn.FacetGrid.html#seaborn-facetgrid>`_
        :type facet_kws: dict[str, Any], Optional
        """
        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.hue = hue
        self.row = row
        self.col = col
        self.palette = palette
        self.kind = kind
        self.legend = legend
        self.plot_kws = plot_kws if plot_kws is not None else {}
        self.facet_kws = facet_kws

    def plot(self) -> None:
        """Implements the plot method of a figure-level relational graph."""
        self.facet_grid = sns.relplot(
            data=self.data_frame,
            x=self.x,
            y=self.y,
            hue=self.hue,
            row=self.row,
            col=self.col,
            palette=self.palette,
            kind=self.kind,
            legend=self.legend,
            facet_kws=self.facet_kws,
            **self.plot_kws,
        )


class DisPlot(FigurePlotter):
    """
    The distribution plot mixin.

    This mixin is the MAFw implementation of the `seaborn displot
    <https://seaborn.pydata.org/generated/seaborn.displot.html#seaborn.displot>`_ and will produce one of the following figure level plots:

    * **histplot**: a simple `histogram
      plot <https://seaborn.pydata.org/generated/seaborn.histplot.html#seaborn.histplot>`_

    * **kdeplot**: a `kernel density <https://seaborn.pydata.org/generated/seaborn.kdeplot.html#seaborn.kdeplot>`_
      estimate plot

    * **ecdfplot**: an `empirical cumulative distribution functions
      <https://seaborn.pydata.org/generated/seaborn.ecdfplot.html#seaborn.ecdfplot>`_  plot

    * **rugplot**: a plot of the `marginal distributions
      <https://seaborn.pydata.org/generated/seaborn.rugplot.html#seaborn.rugplot>`_ as ticks.
    """

    def __init__(
        self,
        x: ColumnName | Iterable[float | complex | int] | None = None,
        y: ColumnName | Iterable[float | complex | int] | None = None,
        hue: ColumnName | Iterable[float | complex | int] | None = None,
        row: ColumnName | Iterable[float | complex | int] | None = None,
        col: ColumnName | Iterable[float | complex | int] | None = None,
        palette: _Palette | matplotlib.colors.Colormap | None = None,
        kind: typing.Literal['hist', 'kde', 'ecdf'] = 'hist',
        legend: bool = True,
        rug: bool = False,
        rug_kws: dict[str, Any] | None = None,
        plot_kws: Mapping[str, Any] | None = None,
        facet_kws: dict[str, Any] | None = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Constructor parameters:

        :param x: The name of the x variable or an iterable containing the x values.
        :type x: str | Iterable, Optional
        :param y: The name of the y variable or an iterable containing the y values.
        :type y: str | Iterable, Optional
        :param hue: The name of the hue variable or an iterable containing the hue values.
        :type hue: str | Iterable, Optional
        :param row: The name of the row category or an iterable containing the row values.
        :type row: str | Iterable, Optional
        :param col: The name of the column category or an iterable containing the column values.
        :type col: str | Iterable, Optional
        :param palette: The colour palette to be used.
        :type palette: str | Colormap, Optional
        :param kind: The type of distribution plot (hist, kde or ecdf). Defaults to hist.
        :type kind: str, Optional
        :param legend: If false, suppress the legend for the semantic variables. Defaults to True.
        :type legend: bool, Optional
        :param rug: If true, show each observation with marginal ticks. Defaults to False.
        :type rug: bool, Optional
        :param rug_kws: Parameters to control the appearance of the rug plot.
        :type rug_kws: Mapping[str, Any], Optional
        :param plot_kws: Parameters passed to the underlying plotting object.
        :type plot_kws: Mapping[str, Any], Optional
        :param facet_kws: Parameters passed to the facet grid object.
        :type facet_kws: Mapping[str, Any], Optional
        """
        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.hue = hue
        self.row = row
        self.col = col
        self.palette = palette
        self.kind = kind
        self.legend = legend
        self.rug = rug
        self.rug_kws = rug_kws
        self.plot_kws = plot_kws if plot_kws is not None else {}
        self.facet_kws = facet_kws

    def plot(self) -> None:
        """Implements the plot method for a figure-level distribution graph"""
        self.facet_grid = sns.displot(
            data=self.data_frame,
            x=self.x,
            y=self.y,
            hue=self.hue,
            row=self.row,
            col=self.col,
            palette=self.palette,
            kind=self.kind,
            legend=self.legend,
            rug=self.rug,
            rug_kws=self.rug_kws,
            facet_kws=self.facet_kws,
            **self.plot_kws,
        )


class CatPlot(FigurePlotter):
    """
    The categorical plot mixin.

    This mixin will produce a figure level categorical plot as described `here
    <https://seaborn.pydata.org/generated/seaborn.catplot.html>`_.

    .. note:

        By default this function treats one of the variables (typically x) as categorical, this means that even if
        this variable is numeric, its value will not be considered. If you want to use the actual value of this
        categorical variable, set native_scale = True.
    """

    def __init__(
        self,
        x: ColumnName | Iterable[float | complex | int] | None = None,
        y: ColumnName | Iterable[float | complex | int] | None = None,
        hue: ColumnName | Iterable[float | complex | int] | None = None,
        row: ColumnName | Iterable[float | complex | int] | None = None,
        col: ColumnName | Iterable[float | complex | int] | None = None,
        palette: _Palette | None = None,
        kind: typing.Literal['strip', 'swarm', 'box', 'violin', 'boxen', 'point', 'bar', 'count'] = 'strip',
        legend: typing.Literal['auto', 'brief', 'full'] | bool = 'auto',
        native_scale: bool = False,
        plot_kws: Mapping[str, Any] | None = None,
        facet_kws: dict[str, Any] | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Constructor parameters:

        :param x: The name of the x variable or an iterable containing the x values.
        :type x: str | Iterable, Optional
        :param y: The name of the y variable or an iterable containing the y values.
        :type y: str | Iterable, Optional
        :param hue: The name of the hue variable or an iterable containing the hue values.
        :type hue: str | Iterable, Optional
        :param row: The name of the row category or an iterable containing the row values.
        :type row: str | Iterable, Optional
        :param col: The name of the column category or an iterable containing the column values.
        :type col: str | Iterable, Optional
        :param palette: The colour palette to be used.
        :type palette: str, Optional
        :param kind: The type of relational plot (scatter or line). Defaults to scatter.
        :type kind: str, Optional
        :param legend: How to draw the legend. If “brief”, numeric hue and size variables will be represented with a
            sample of evenly spaced values. If “full”, every group will get an entry in the legend.
            If “auto”, choose between brief or full representation based on number of levels.
            If False, no legend data is added and no legend is drawn. Defaults to auto.
        :type legend: str | bool, Optional
        :param native_scale: When True, numeric or datetime values on the categorical axis will maintain their original
            scaling rather than being converted to fixed indices. Defaults to False.
        :type native_scale: bool, Optional
        :param plot_kws: A dictionary like list of keywords passed to the underlying `seaborn.catplot
            <https://seaborn.pydata.org/generated/seaborn.relplot.html#seaborn.catplot>`_.
        :type plot_kws: dict[str, Any], Optional
        :param facet_kws: A dictionary like list of keywords passed to the underlying
            `seaborn.FacetGrid <https://seaborn.pydata.org/generated/seaborn.FacetGrid.html#seaborn-facetgrid>`_
        :type facet_kws: dict[str, Any], Optional
        """
        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.hue = hue
        self.row = row
        self.col = col
        self.palette = palette
        self.kind = kind
        self.legend = legend
        self.native_scale = native_scale
        self.plot_kws = plot_kws if plot_kws is not None else {}
        self.facet_kws = facet_kws

    def plot(self) -> None:
        """Implements the plot method of a figure-level categorical graph."""
        self.facet_grid = sns.catplot(
            data=self.data_frame,
            x=self.x,
            y=self.y,
            hue=self.hue,
            row=self.row,
            col=self.col,
            palette=self.palette,
            kind=self.kind,
            legend=self.legend,
            native_scale=self.native_scale,
            facet_kws=self.facet_kws,
            **self.plot_kws,
        )


@processor_depends_on_optional(module_name='pandas;seaborn', warn=True, raise_ex=False)
class GenericPlotter(Processor, metaclass=PlotterMeta):
    """
    The Generic Plotter processor.

    This is a subclass of a Processor with advanced functionality to fetch data in the form of a dataframe and to
    produce plots.

    The key difference with respect to a normal processor is it :meth:`.process` method that has been already
    implemented as follows:

    .. literalinclude:: ../../../src/mafw/processor_library/plotter.py
        :pyobject: GenericPlotter.process
        :dedent:

    This actually means that when you are subclassing a GenericPlotter you do not have to implement the process method
    as you would do for a normal Processor, but you will have to implement the following methods:

        * :meth:`~.in_loop_customization`.

            The processor execution workflow (LoopType) can be any of the available, so
            actually the process method might be invoked only once, or multiple times inside a loop structure
            (for or while).
            If the execution is cyclic, then you may want to have the possibility to do some customisation for each
            iteration, for example, changing the plot title, or modifying the data selection, or the filename where the
            plots will be saved.

            You can use this method also in case of a single loop processor, in this case you will not have access to
            the loop parameters.

        * :meth:`~.get_data_frame`.

            This method has the task to get the data to be plotted in the form of a pandas DataFrame. The processor has
            the :attr:`~.data_frame` attribute where the data will be stored to make them accessible from all other
            methods.

        * :meth:`~.patch_data_frame`.

            A convenient method to apply data frame manipulation to the data just retrieved.

        * :meth:`~.plot`.

            This method is where the real plotting is taking place. Use the :attr:`~.data_frame` to plot the quantities
            you want.

        * :meth:`~.customize_plot`.

            This method can be optionally used to customise the appearance of the facet grid produced by the
            :meth:`~plot` method. It is particularly useful when the user is mixing this class with one of the
            :class:`~.FigurePlotter` mixin, thus not having direct access to the plot method.

        * :meth:`~.save`.

            This method is where the produced plot is saved in a file. Remember to append the output file name to the
            :attr:`list of produced outputs <.output_filename_list>` so that the :meth:`~._update_plotter_db` method
            will automatically store this file in the database during the :meth:`~.finish` execution.

        * :meth:`~.update_db`.

            If the user wants to update a specific table in the database, they can use this method.

            It is worth reminding that all plotters are saving all generated files in the standard table PlotterOutput.
            This is automatically done by the :meth:`~._update_plotter_db` method that is called in the
            :meth:`~.finish` method.

    You do not need to overload the :meth:`~.slice_data_frame` nor the :meth:`~.group_and_aggregate_data_frame`
    methods, but you can simply use them by setting the :attr:`~.slicing_dict` and the :attr:`~.grouping_columns`
    and the :attr:`~.aggregation_functions`.
    """

    def __init__(
        self,
        slicing_dict: MutableMapping[str, Any] | None = None,
        grouping_columns: Iterable[str] | None = None,
        aggregation_functions: Iterable[str | Callable[[Any], Any]] | None = None,
        matplotlib_backend: str = 'Agg',
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Constructor parameters:

        :param slicing_dict: A dictionary with key, value pairs to slice the input data frame before the plotting
            occurs.
        :type slicing_dict: dict[str, Any], Optional
        :param grouping_columns: A list of columns for the groupby operation on the data frame.
        :type grouping_columns: list[str], Optional
        :param aggregation_functions: A list of functions for the aggregation on the grouped data frame.
        :type aggregation_functions: list[str | Callable[[Any], Any], Optional
        :param matplotlib_backend: The name of the matplotlib backend to be used. Defaults to 'Agg'
        :type matplotlib_backend: str, Optional
        """
        super().__init__(*args, **kwargs)

        # attributes that can be set in the constructor

        self.slicing_dict: MutableMapping[str, Any] | None = slicing_dict
        """The dictionary for slicing the input data frame"""

        self.grouping_columns: Iterable[str] | None = grouping_columns
        """The list of columns for grouping the data frame"""

        self.aggregation_functions: Iterable[str | Callable[[Any], Any]] | None = aggregation_functions
        """The list of aggregation functions to be applied to the grouped dataframe"""

        self.matplotlib_backend: str = matplotlib_backend
        """The backend to be used for matplotlib."""

        # internal use attributes.

        self.data_frame: pd.DataFrame = pd.DataFrame()
        """The pandas DataFrame containing the data to be plotted."""

        self.output_filename_list: list[Path] = []
        """The list of produced filenames."""

        self.facet_grid: sns.FacetGrid | None = None
        """The reference to the facet grid."""

        # private attributes

        # be sure that some additional methods if implemented are calling the super.
        self._methods_to_be_checked_for_super.extend([('patch_data_frame', GenericPlotter)])  # type: ignore[list-item]

    @typing.no_type_check
    def is_output_existing(self) -> bool:
        """
        Check for plotter output existence.

        Generally, plotter subclasses do not have a real output that can be saved to a database. This class is meant to
        generate one or more graphical output file.

        One of the biggest advantages of having the output of a processor stored in the database is the ability to
        conditionally execute the processor if, and only if, the output is missing or changed.

        In order to allow also plotter processor to benefit from this feature, a :class:`dedicated table
        <.PlotterOutput>` is available among the :ref:`standard tables <std_tables>`.

        If a connection to the database is provided, then this method is invoked at the beginning of the
        :meth:`~.process` and a select query over the :class:`~.PlotterOutput` model is executed filtering by
        processor name. All files in the filename lists are checked for existence and also the checksum is verified.

        This method will return True, if the output of the processor is already existing and valid, False, otherwise.

        :return: True if the processor output exists and it is valid.
        :rtype: bool
        """
        if self._database is None:
            # no active database connection. it makes no sense to continue. inform the user and return
            log.warning('No database connection available. Impossible to check for existing output')
            return False

        try:
            query = PlotterOutput.get(PlotterOutput.plotter_name == self.name)
            # check if all files exist:
            if not all([f.exists() for f in query.filename_list]):
                # at least one file is missing.
                # delete the whole row and continue
                PlotterOutput.delete().where(PlotterOutput.plotter_name == self.name).execute()
                return False
            else:
                # all files exist.
                # check that they are still actual
                if query.checksum != file_checksum(query.filename_list):
                    # at least one file is changed.
                    # delete the whole row and continue
                    PlotterOutput.delete().where(PlotterOutput.plotter_name == self.name).execute()
                    return False
                else:
                    # all files exit and the checksum is the same.
                    # we stop it here
                    return True

        except peewee.DoesNotExist:
            # no output for this plotter processor found in the DB.
            return False

    def start(self) -> None:
        """
        Overload of the start method.

        The :class:`~.GenericPlotter` is overloading the :meth:`~.Processor.start` in order to allow the user to
        change the matplotlib backend.

        The user can selected which backend to use either directly in the class constructor or assign it to the class
        attribute :attr:`~.matplotlib_backend`.
        """
        super().start()
        try:
            if plt.get_backend() != self.matplotlib_backend:
                plt.switch_backend(self.matplotlib_backend)
        except ModuleNotFoundError:
            log.critical('%s is not a valid plt backend' % self.matplotlib_backend)
            raise

    def process(self) -> None:
        """
        Process method overload.

        In the case of a plotter subclass, the process method is already implemented and the user should not overload
        it. On the contrary, the user must overload the other implementation methods described in the general
        :class:`class description <.GenericPlotter>`.
        """
        if self.filter_register.new_only:
            if self.is_output_existing():
                return

        self.in_loop_customization()
        self.get_data_frame()
        self.patch_data_frame()
        self.slice_data_frame()
        self.group_and_aggregate_data_frame()
        if len(self.data_frame.index):
            self.plot()
            self.customize_plot()
            self.save()
            self.update_db()
            plt.close('all')

    def in_loop_customization(self) -> None:
        """
        Customize the parameters for the output or input data for each execution iteration.
        """
        pass

    def get_data_frame(self) -> None:
        """
        Get the data frame with the data to be plotted.

        This method can be either implemented in the GenericPlotter subclass or via a :class:`.DataRetriever` mixin
        class.
        """
        # it must be overloaded.
        pass

    def format_progress_message(self) -> None:
        self.progress_message = f'{self.name} is working'

    def plot(self) -> None:
        """
        The plot method.

        This is where the user has to implement the real plot generation
        """
        pass

    def customize_plot(self) -> None:
        """
        The customise plot method.

        The user can overload this method to customise the output produced by the :meth:`~.plot` method, like for
        example adding meaningful axis titles, changing format, and so on.

        As usual, it is possible to use the :attr:`~.item`, :attr:`~.i_item` and :attr:`~.n_item` to access the loop
        parameters, :attr:`~.facet_grid` to access the seaborn FacetGrid object produced in the :meth:`~.plot` method
        and also the :attr:`~.data_frame` to retrieve information directly from the data set.
        """
        pass

    def save(self) -> None:
        """
        The save method.

        This is where the user has to implement the saving of the plot on disc.
        """
        pass

    def update_db(self) -> None:
        """
        The update database method.

        This is where the user has to implement the optional update of the database.

        .. seealso:

            The plotter output table is automatically update by :meth:`~._update_plotter_db`.
        """
        pass

    def slice_data_frame(self) -> None:
        """
        Perform data frame slicing

        The user can set some slicing criteria in the :attr:`~.slicing_dict` to select some specific data subset. The
        values of the slicing dict can be changed during each iteration within the implementation of the
        :meth:`~.in_loop_customization`.

        .. seealso::
            This method is simply invoking the :func:`~.slice_data_frame` function from the :mod:`~.pandas_tools`.
        """
        if self.slicing_dict:
            self.data_frame = slice_data_frame(self.data_frame, self.slicing_dict)

    def group_and_aggregate_data_frame(self) -> None:
        """
        Performs groupyby and aggregation of the data frame.

        If the user provided some :attr:`grouping columns <.grouping_columns>` and :attr:`aggregation functions
        <.aggregation_functions>` then the :func:`~.group_and_aggregate_data_frame` is invoked accordingly.

        The user can update the values of those attributes during each cycle iteration within the implementation of
        the :meth:`~.in_loop_customization`.

        .. seealso::
            This method is simply invoking the :func:`~.group_and_aggregate_data_frame` function from the :mod:`~.pandas_tools`.
        """
        if self.grouping_columns and self.aggregation_functions:
            self.data_frame = group_and_aggregate_data_frame(
                self.data_frame, self.grouping_columns, self.aggregation_functions
            )

    def finish(self) -> None:
        if self.looping_status == LoopingStatus.Continue:
            self._update_plotter_db()  # type: ignore[no-untyped-call]
        super().finish()

    def patch_data_frame(self) -> None:
        """
        Modify the data frame

        This method can be used to perform operation on the data frame, like adding new columns.
        It can be either implemented in the plotter processor subclasses or via a mixin class.
        """
        pass

    @typing.no_type_check
    def _update_plotter_db(self) -> None:
        """
        Updates the Plotter DB.

        A plotter subclass is producing in most of the cases just a plot as an output, so no additional information
        should be stored in the database, but this is already enough to avoid executing the processor if not needed.

        This method is actually protected against execution without a valid database instance.
        """
        if self._database is None:
            # there is no active database connection. No need to continue. Inform the user and continue
            log.warning('No database connection available. Impossible to update the plotter output')
            return

        if len(self.output_filename_list) == 0:
            # there is no need to make an entry because there are no saved file
            return

        PlotterOutput.std_upsert(
            {
                'plotter_name': self.name,
                'filename_list': self.output_filename_list,
                'checksum': self.output_filename_list,
            }
        ).execute()



"""
Module provides standard tables that are included in all database created by MAFw processor.

Standard tables are automatically created and initialized by a :class:`~mafw.processor.Processor` or a
:class:`~mafw.processor.ProcessorList` when opening a database connection.

This means that if a processor receives a valid database object, then it will suppose that the connection was already
opened somewhere else (either from a ProcessorList or a third party) and thus it is not creating the standard tables.

If a processor is constructed using a database configuration dictionary, then it will first try to open a connection
to the DB, then creating all standard tables and finally executing their :class:`StandardTable.init` method. The same
apply for the Processor list.

In other words, object responsible to open the database connection is taking care also of creating the standard
tables and of initializing them. If the user opens the connection and pass it to a Processor or ProcessorList,
then the user is responsible to create the standard tables and to initialise them.

All standard tables must derive from the :class:`StandardTable` in order to have the same interface for the
initialisation.

Users can create their own standard tables and export them as pluggable object using the same approach as for custom
processors. An example is provided in the :ref:`documentation <my_std_tables.py>`.

"""

import peewee
from peewee import AutoField, BooleanField, TextField

from mafw.db.db_model import MAFwBaseModel
from mafw.db.fields import FileChecksumField, FileNameListField
from mafw.db.trigger import Trigger, TriggerAction, TriggerWhen


class StandardTable(MAFwBaseModel):
    """A base class for tables that are generated automatically by the MAFw processor."""

    @classmethod
    def init(cls) -> None:
        """The user must overload this method, if he wants some specific operations to be performed on the model
        everytime the database is connected."""
        pass


class StandardTableDoesNotExist(Exception):
    """An exception raised when trying to access a not existing table."""


class TriggerStatus(StandardTable):
    """A Model for the trigger status"""

    trigger_type_id = AutoField(primary_key=True, help_text='Primary key')
    trigger_type = TextField(
        help_text='You can use it to specify the type (DELETE/INSERT/UPDATE) or the name of a specific trigger'
    )
    status = BooleanField(default=True, help_text='False (0) = disable / True (1) = enable')

    @classmethod
    def init(cls) -> None:
        """Resets all triggers to enable when the database connection is opened."""
        data = [
            dict(trigger_type_id=1, trigger_type='DELETE', status=True),
            dict(trigger_type_id=2, trigger_type='INSERT', status=True),
            dict(trigger_type_id=3, trigger_type='UPDATE', status=True),
            dict(trigger_type_id=4, trigger_type='DELETE_FILES', status=True),
        ]
        cls.insert_many(data).on_conflict_replace().execute()


class TriggerStatusDoesNotExist(Exception):
    """An exception raised when trying to access a not existing table."""


class OrphanFile(StandardTable):
    """A Model for the files to be removed from disc"""

    file_id = AutoField(primary_key=True, help_text='Primary key')
    filenames = FileNameListField(help_text='The path to the file to be deleted', checksum_field='checksum')
    checksum = FileChecksumField(help_text='The checksum of the files in the list.')


class OrphanFileDoesNotExist(peewee.DoesNotExist):
    """An exception raised when trying to access a not existing table."""


class PlotterOutput(StandardTable):
    """
    A model for the output of the plotter processors.

    The model has a trigger activated on delete queries to insert filenames and checksum in the OrphanFile model.
    """

    plotter_name = TextField(primary_key=True, help_text='The plotter processor name')
    filename_list = FileNameListField(help_text='The path to the output file', checksum_field='checksum')
    checksum = FileChecksumField(help_text='The checksum of the files in the list.')

    @classmethod
    def triggers(cls) -> list[Trigger]:
        insert_into_orphan = Trigger('plotter_delete_file', (TriggerWhen.Before, TriggerAction.Delete), cls, safe=True)
        # warning:
        #   we cannot use the filenames for both the filenames and checksum fields because it is a bare sql statement
        insert_into_orphan.add_sql(
            'INSERT INTO orphan_file (filenames, checksum) VALUES (OLD.filename_list, OLD.checksum);'
        )
        return [insert_into_orphan]

    class Meta:
        depends_on = [OrphanFile]


class PlotterOutputDoesNotExist(peewee.DoesNotExist):
    """An exception raised when trying to access a not existing table."""


standard_tables: dict[str, type[StandardTable]] = {}
"""
A dictionary containing the standard tables being exported to the rest of the framework. 

The key is the name of the model and the value is the model class itself.
"""

standard_tables.update({'TriggerStatus': TriggerStatus, 'OrphanFile': OrphanFile, 'PlotterOutput': PlotterOutput})



import pytest

from mafw.active import Active


class MyClass:
    status = Active('OK')

    def do(self):
        self.status = 'ERROR'

    @staticmethod
    def on_status_change(old_status, new_status):
        if new_status == 'ERROR':
            raise Exception(f'MyClass status is {new_status}')


def test_action_functionality():
    with pytest.raises(Exception) as e:
        my_class = MyClass()
        my_class.do()

    assert 'ERROR' in str(e)



from pathlib import Path

import pytest
from pluggy import PluginManager

import mafw.mafw_errors
from mafw.runner import MAFwApplication
from mafw.ui.abstract_user_interface import UserInterfaceBase


def test_application_creation1(not_existing_steering_file):
    # creation of an app with a not existing steering file
    with pytest.raises(FileNotFoundError):
        MAFwApplication(not_existing_steering_file)


def test_application_creation2(invalid_steering_file):
    # creation of an app with an invalid steering file and no other arguments.
    with pytest.raises(mafw.mafw_errors.InvalidSteeringFile):
        MAFwApplication(invalid_steering_file)


def test_application_creation3(valid_steering_file):
    # creation of an app with valid steering file and no other arguments.
    app = MAFwApplication(valid_steering_file)
    assert isinstance(app.steering_file, Path)
    assert isinstance(app.plugin_manager, PluginManager)
    assert isinstance(app.user_interface, UserInterfaceBase)
    assert app.user_interface.name == 'rich'


def test_application_creation4(valid_steering_file):
    # creation of an app with valid steering file as string
    steering_file_name_str = str(valid_steering_file)
    app = MAFwApplication(steering_file_name_str)
    assert isinstance(app.steering_file, Path)


def test_application_creation5(valid_steering_file):
    # creation of an app with valid steering file and user interface set to rich
    # but overridden to console in the constructor.
    app = MAFwApplication(valid_steering_file, user_interface='console')
    assert isinstance(app.user_interface, UserInterfaceBase)
    assert app.user_interface.name == 'console'


def test_application_creation6(valid_steering_file):
    # creation of an app with valid steering file and user interface set to rich
    # but overridden to something wrong in the constructor.
    # still get console has fall back.
    app = MAFwApplication(valid_steering_file, user_interface='unknown')
    assert isinstance(app.user_interface, UserInterfaceBase)
    assert app.user_interface.name == 'console'


def test_application_creation7(valid_steering_file):
    # creation of an app with no steering file
    app = MAFwApplication()
    assert not app._initialized

    # attempt to run a not initialised app.
    with pytest.raises(mafw.mafw_errors.RunnerNotInitialized):
        app.run()

    # attempt to run a not initialised app passing steering file to run.
    app.run(valid_steering_file)
    assert app._initialized



from pathlib import Path, PosixPath, WindowsPath

import pytest
import tomlkit.exceptions
from tomlkit.toml_file import TOMLFile

import mafw.mafw_errors
import mafw.tools.toml_tools
from mafw.db.db_configurations import default_conf
from mafw.decorators import single_loop
from mafw.processor import ActiveParameter, PassiveParameter, Processor
from mafw.tools.toml_tools import (
    _add_db_configuration,
    dump_processor_parameters_to_toml,
    generate_steering_file,
    load_steering_file,
)


@single_loop
class ActiveParameterProcessor(Processor):
    """A processor with one active parameter."""

    active_param = ActiveParameter('active_param', default=-1, help_doc='An active parameter with default value -1')


@single_loop
class AnotherActiveParameterProcessor(Processor):
    """Another processor with one active parameter."""

    active_param = ActiveParameter('active_param', default=-1, help_doc='An active parameter with default value -1')


class PassiveParameterProcessor(Processor):
    """A processor with one passive parameter."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, looper='single', **kwargs)
        self.passive_param = PassiveParameter(
            'passive_param', default=-1, help_doc='A passive parameter with default value -1'
        )


class AnotherPassiveParameterProcessor(Processor):
    """Another processor with one passive parameter."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, looper='single', **kwargs)
        self.passive_param = PassiveParameter(
            'passive_param', default=-1, help_doc='A passive parameter with default value -1'
        )


@pytest.fixture
def registered_example_processor() -> list[type[Processor]]:
    from mafw.examples.loop_modifier import ModifyLoopProcessor
    from mafw.examples.sum_processor import AccumulatorProcessor, GaussAdder

    return [AccumulatorProcessor, GaussAdder, ModifyLoopProcessor]


def get_toml(input_file: str | Path) -> dict:
    return TOMLFile(input_file).read().value


def test_wrong_type():
    output_file_example = 'test1.toml'
    with pytest.raises(TypeError):
        processor_list = [
            ActiveParameterProcessor,
            ActiveParameterProcessor,
            'a string',
            PassiveParameterProcessor,
            AnotherActiveParameterProcessor,
        ]
        dump_processor_parameters_to_toml(processor_list, output_file_example)

    Path(output_file_example).unlink(missing_ok=True)


def test_same_name_exception():
    output_file_example = 'test1.toml'
    with pytest.raises(tomlkit.exceptions.KeyAlreadyPresent) as e:
        processor_list = [
            ActiveParameterProcessor,
            ActiveParameterProcessor,
            PassiveParameterProcessor,
            AnotherActiveParameterProcessor,
        ]
        dump_processor_parameters_to_toml(processor_list, output_file_example)
    assert 'ActiveParameterProcessor' in str(e.value)
    Path(output_file_example).unlink(missing_ok=True)


def test_dump_processor_parameters_to_toml_1(datadir, tmp_path):
    output_file_example = 'test1.toml'

    processor_list = [
        ActiveParameterProcessor,
        AnotherPassiveParameterProcessor,
        PassiveParameterProcessor,
        AnotherActiveParameterProcessor,
    ]

    dump_processor_parameters_to_toml(processor_list, tmp_path / output_file_example)

    expected_output_file = datadir / f'expected_{output_file_example}'
    expected_dict = get_toml(expected_output_file)
    obtained_dict = get_toml(tmp_path / output_file_example)

    # the dictionaries are as expected
    assert expected_dict == obtained_dict

    # the text file is identical
    with open(tmp_path / output_file_example) as fp:
        generated_file = fp.readlines()

    with open(expected_output_file) as fp:
        expected_file = fp.readlines()

    # skip the first line because there is timestamp
    assert generated_file[1:] == expected_file[1:]


def test_steering_file_generation_with_no_db(datadir, tmp_path, registered_example_processor):
    output_file_example = 'steering-file.toml'
    generate_steering_file(tmp_path / output_file_example, registered_example_processor)

    expected_output_file = datadir / f'generated-{output_file_example}'
    expected_dict = get_toml(expected_output_file)
    obtained_dict = get_toml(tmp_path / output_file_example)

    # the dictionaries should have the same keys
    assert list(expected_dict.keys()) == list(obtained_dict.keys())
    # values might be different depending on previous tests.


def test_dump_processor_parameters_to_toml_2(datadir, tmp_path):
    output_file_example = 'test2.toml'

    # create an instance of ActiveParameterProcessor with a specific value of the parameter
    # but we will include the class in the processor list.
    ActiveParameterProcessor(active_param=100)

    # create an instance of AnotherActiveParameterProcessor with a specific value of the parameter
    # and we will submit the instance
    another_active_processor_instance = AnotherActiveParameterProcessor(active_param=101)

    # create an instance of PassiveParameterProcessor with a specific value of the parameter
    # but we will submit the class via the use of type.
    passive_processor_instance = PassiveParameterProcessor(passive_param=102)

    # create an instance of AnotherPassiveParameterProcessor with a specific value of the parameter
    # and we will submit the instance.
    another_passive_processor_instance = AnotherPassiveParameterProcessor(passive_param=103)

    processor_list = [
        ActiveParameterProcessor,  # a class
        another_active_processor_instance,  # an instance
        type(passive_processor_instance),  # a class
        another_passive_processor_instance,  # an instance
    ]
    dump_processor_parameters_to_toml(processor_list, tmp_path / output_file_example)
    expected_output_file = datadir / f'expected_{output_file_example}'
    expected_dict = get_toml(expected_output_file)
    obtained_dict = get_toml(tmp_path / output_file_example)

    # the dictionaries are as expected
    assert expected_dict == obtained_dict

    # the text file is identical
    with open(tmp_path / output_file_example) as fp:
        generated_file = fp.readlines()

    with open(expected_output_file) as fp:
        expected_file = fp.readlines()

    # skip the first line because there is timestamp
    assert generated_file[1:] == expected_file[1:]


def test_load_steering_file():
    with pytest.raises(FileNotFoundError) as e:
        load_steering_file('pippo')
    assert 'pippo' in str(e.value)


def test_validating_steering_file(datadir):
    valid_file = 'valid-steering-file.toml'
    load_steering_file(datadir / valid_file)

    invalid_files = range(1, 3)
    for i in invalid_files:
        invalid_file = f'invalid-steering-file{i}.toml'
        with pytest.raises(mafw.mafw_errors.InvalidSteeringFile):
            load_steering_file(datadir / invalid_file)


def test_add_db_configuration(caplog):
    # test exception raise with no conf and unknown engine
    with pytest.raises(mafw.mafw_errors.UnknownDBEngine) as e:
        _add_db_configuration(db_engine='unknown')
    assert 'unknown' in str(e.value)
    assert '(unknown) is not yet implemented' in caplog.text
    caplog.clear()

    # test exception raise with unknown engine and invalid configuration
    conf = {'invalid': 'invalid'}
    with pytest.raises(mafw.mafw_errors.UnknownDBEngine) as e:
        _add_db_configuration(database_conf=conf, db_engine='unknown')
    assert len(caplog.records) == 2
    assert 'The provided database configuration is invalid.' in caplog.text
    assert '(unknown) is not yet implemented' in caplog.text
    caplog.clear()

    # test log message with known engine and invalid configuration
    engine = 'mysql'
    doc = _add_db_configuration(database_conf=conf, db_engine=engine)
    assert 'The provided database configuration is invalid. Adding default configuration' in caplog.text
    assert doc.value['DBConfiguration'] == default_conf[engine]
    caplog.clear()

    # test log message with invalid configuration and no explict db_engine
    doc = _add_db_configuration(database_conf=conf)
    assert 'The provided database configuration is invalid. Adding default configuration' in caplog.text
    assert doc.value['DBConfiguration'] == default_conf['sqlite']
    caplog.clear()


def test_path_encoding():
    obj = {'path': Path.cwd()}
    doc = tomlkit.dumps(obj)
    if isinstance(Path.cwd(), WindowsPath):
        assert doc == f"""path = '{str(Path.cwd())}'\n"""
    elif isinstance(Path, PosixPath):
        assert doc == f"""path = "{str(Path.cwd())}"\n"""
